@@ -25,9 +25,11 @@
 #include <stdint.h>
 #include <limits.h>
 #include <sys/stat.h>
+#include <sys/vfs.h>
 #include <sys/file.h>
 #include <sys/wait.h>
 #include <sys/user.h>
+#include <sys/param.h>
 #include <sys/ioctl.h>
 #include <linux/types.h>
 #include <linux/fs.h>
@@ -75,16 +77,6 @@ static int open_device(const char *device)
 }
 
 
-#define ioctl_device(fd, req, arg)					\
-	({								\
-		int __ret = 0;						\
-		if (ioctl(fd, req, arg)) {				\
-			ploop_err(errno, "Error in ioctl(" #req ")");	\
-			__ret = SYSEXIT_DEVIOC;				\
-		}							\
-		__ret;							\
-	 })
-
 static int fsync_balloon(int fd)
 {
 	if (fsync(fd)) {
@@ -393,7 +385,7 @@ int ploop_balloon_change_size(const char *device, int balloonfd, off_t new_size)
 		goto err;
 	if (entries_used == 0) {
 		drop_state = 1;
-		ploop_log(0, "no unused cluster blocks found");
+		ploop_log(0, "No unused cluster blocks found");
 		goto out;
 	}
 
@@ -431,7 +423,7 @@ out:
 err:
 	if (drop_state) {
 		memset(&b_ctl, 0, sizeof(b_ctl));
-		ioctl(fd, PLOOP_IOC_BALLOON, &b_ctl);
+		(void)ioctl_device(fd, PLOOP_IOC_BALLOON, &b_ctl);
 	}
 	close(fd);
 	free(pfiemap);
@@ -447,7 +439,7 @@ err:
 	return ret;
 }
 
-int ploop_balloon_get_state(const char *device, __u32  *state)
+int ploop_balloon_get_state(const char *device, __u32 *state)
 {
 	int fd, ret;
 	struct ploop_balloon_ctl b_ctl;
@@ -601,7 +593,7 @@ int ploop_balloon_complete(const char *device)
 	if (fd == -1)
 		return -1;
 
-	err = ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL);
+	err = ioctl(fd, PLOOP_IOC_DISCARD_FINI, NULL);
 	if (err && errno != EBUSY) {
 		ploop_err(errno, "Can't finalize discard mode");
 		goto out;
@@ -638,7 +630,7 @@ out:
 	return err;
 }
 
-int ploop_balloon_check_and_repair(const char *device, char *mount_point, int repair)
+int ploop_balloon_check_and_repair(const char *device, const char *mount_point, int repair)
 {
 	int   ret, fd = -1;
 	int   balloonfd = -1;
@@ -803,7 +795,7 @@ int ploop_balloon_check_and_repair(const char *device, char *mount_point, int re
 err:
 	if (drop_state) {
 		memset(&b_ctl, 0, sizeof(b_ctl));
-		ioctl(fd, PLOOP_IOC_BALLOON, &b_ctl);
+		(void)ioctl_device(fd, PLOOP_IOC_BALLOON, &b_ctl);
 	}
 
 	// FIXME: close_delta()
@@ -825,10 +817,13 @@ static void stop_trim_handler(int sig)
 	trim_stop = 1;
 }
 
-static int ploop_trim(const char *mount_point, __u64 minlen_b)
+/* The fragmentation of such blocks doesn't affect the speed of w/r */
+#define MAX_DISCARD_CLU 32
+
+static int ploop_trim(const char *mount_point, __u64 minlen_b, __u64 cluster)
 {
-	struct fstrim_range range = {0, ULLONG_MAX, minlen_b};
-	int fd, ret;
+	struct fstrim_range range = {0, ULLONG_MAX, 0};
+	int fd, ret = -1;
 
 	struct sigaction sa = {
 		.sa_handler     = stop_trim_handler,
@@ -848,12 +843,26 @@ static int ploop_trim(const char *mount_point, __u64 minlen_b)
 
 	sys_syncfs(fd);
 
-	ret = ioctl(fd, FITRIM, &range);
-	if (ret < 0) {
-		if (trim_stop)
-			ret = 0;
-		else
-			ploop_err(errno, "Can't trim file system");
+	if (minlen_b < cluster)
+		minlen_b = cluster;
+	else
+		minlen_b = (minlen_b + cluster - 1) / cluster * cluster;
+	range.minlen = MAX(MAX_DISCARD_CLU * cluster, minlen_b);
+
+	for (; range.minlen >= minlen_b; range.minlen /= 2) {
+		ploop_log(1, "Call FITRIM, for minlen=%lld", range.minlen);
+		ret = ioctl(fd, FITRIM, &range);
+		if (ret < 0) {
+			if (trim_stop)
+				ret = 0;
+			else
+				ploop_err(errno, "Can't trim file system");
+			break;
+		}
+
+		/* last iteration should go with range.minlen == minlen_b */
+		if (range.minlen != minlen_b && range.minlen / 2 < minlen_b)
+			range.minlen = minlen_b * 2;
 	}
 
 	close(fd);
@@ -861,83 +870,24 @@ static int ploop_trim(const char *mount_point, __u64 minlen_b)
 	return ret;
 }
 
-#define ilog2(n) (			\
-	(n) >= (1 << 9) ? 9 :		\
-	(n) & 1 << 8	 ? 8 :		\
-	(n) & 1 << 7	 ? 7 :		\
-	(n) & 1 << 6	 ? 6 :		\
-	(n) & 1 << 5	 ? 5 :		\
-	(n) & 1 << 4	 ? 4 :		\
-	(n) & 1 << 3	 ? 3 :		\
-	(n) & 1 << 2	 ? 2 :		\
-	(n) & 1 << 1	 ? 1 :	0	\
-)
-
-#define DISCARD_TABLE_SIZE 10
-
-static int discard_collect_stat(int fd, __u32 *distrib)
-{
-	struct ploop_freeblks_ctl  *freeblks = NULL;
-	int ret, i;
-
-	ret = freeblks_alloc(&freeblks, 0);
-	if (ret)
-		return ret;
-
-	ret = ioctl_device(fd, PLOOP_IOC_FBGET, freeblks);
-	if (ret)
-		goto free;
-
-	if (freeblks->n_extents == 0) {
-		// This should never happen, and we don't want
-		// to return an error here.
-		ploop_err(EINVAL, "The number of extents is zero");
-		goto free;
-	}
-
-	ret = freeblks_alloc(&freeblks, freeblks->n_extents);
-	if (ret)
-		goto free;
-
-	ret = ioctl_device(fd, PLOOP_IOC_FBGET, freeblks);
-	if (ret)
-		goto free;
-
-	for (i = 0; i < freeblks->n_extents; i++) {
-		__u32 len = freeblks->extents[i].len;
-		distrib[ilog2(len)] += len;
-	}
-	ret = ioctl_device(fd, PLOOP_IOC_FBDROP, 0);
-	if (ret)
-		goto free;
-free:
-	free(freeblks);
-	return ret;
-}
-
-enum {
-	PLOOP_DISCARD_STAT,
-	PLOOP_DISCARD_COMPACT,
-
-	PLOOP_DISCARD_MAX
-};
-
-static int __ploop_discard(int fd, const char *device, const char *mount_point,
-			int state, __u32 *minlen_c, __u32 cluster, __u32 to_free)
+static int __ploop_discard(struct ploop_disk_images_data *di, int fd,
+			const char *device, const char *mount_point,
+			__u64 minlen_b, __u32 cluster, __u32 to_free,
+			const int *stop)
 {
 	pid_t tpid;
-	int exit_code = 0, ret, status;
-	__u32 distrib[DISCARD_TABLE_SIZE], size = 0;
+	int err = 0, ret, status;
+	__u32 size = 0;
 
-	memset(distrib, 0, sizeof(distrib));
-
-	ploop_log(3, "Trying to find free extents bigger than %u clusters", *minlen_c);
+	ploop_log(3, "Trying to find free extents bigger than %llu bytes", minlen_b);
 
+	if (ploop_lock_di(di))
+		return SYSEXIT_LOCK;
 	ret = ioctl_device(fd, PLOOP_IOC_DISCARD_INIT, NULL);
+	ploop_unlock_di(di);
 	if (ret) {
 		ploop_err(errno, "Can't initialize discard mode");
-		close(fd);
-		return 1;
+		return ret;
 	}
 
 	tpid = fork();
@@ -946,26 +896,21 @@ static int __ploop_discard(int fd, const char *device, const char *mount_point,
 		if (ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL))
 			ploop_err(errno, "Can't finalize discard mode");
 
-		close(fd);
 		return -1;
 	}
 
 	if (tpid == 0) {
-		ret = ploop_trim(mount_point, (__u64) *minlen_c * cluster);
-		if (ret < 0)
-			exit_code = 1;
-
+		ret = ploop_trim(mount_point, minlen_b, cluster);
 		if (ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL))
 			ploop_err(errno, "Can't finalize discard mode");
 
-		close(fd);
-		exit(exit_code);
+		exit(ret < 0);
 	}
 
 	while (1) {
 		struct ploop_balloon_ctl b_ctl;
 
-		ploop_log(0, "Waiting");
+		ploop_log(3, "Waiting");
 		ret = ioctl(fd, PLOOP_IOC_DISCARD_WAIT, NULL);
 		if (ret < 0) {
 			ploop_err(errno, "Waiting for a discard request failed");
@@ -973,7 +918,8 @@ static int __ploop_discard(int fd, const char *device, const char *mount_point,
 		} else if (ret == 0)
 			break;
 
-		ret = ioctl(fd, PLOOP_IOC_FBFILTER, *minlen_c);
+		/* FIXME PLOOP_IOC_DISCARD_WAIT should return size */
+		ret = ioctl(fd, PLOOP_IOC_FBFILTER, 0);
 		if (ret < 0) {
 			ploop_err(errno, "Can't filter free blocks");
 			break;
@@ -985,94 +931,91 @@ static int __ploop_discard(int fd, const char *device, const char *mount_point,
 			continue;
 		} else
 			size += ret;
-
-		switch (state) {
-		case PLOOP_DISCARD_COMPACT:
-			memset(&b_ctl, 0, sizeof(b_ctl));
-			b_ctl.keep_intact = 1;
-			ret = ioctl_device(fd, PLOOP_IOC_BALLOON, &b_ctl);
-			if (ret)
-				break;
-
-			if (b_ctl.mntn_type == PLOOP_MNTN_OFF) {
-				ploop_log(0, "Unexpected maintenance type 0x%x", b_ctl.mntn_type);
-				ret = -1;
-				break;
-			}
-
-			if (size >= to_free) {
-				ploop_log(3, "Killing the trim process %d", tpid);
-				kill(tpid, SIGUSR1);
-				ret = ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL);
-				if (ret < 0 && errno != EBUSY)
-					ploop_err(errno, "Can't finalize a discard mode");
-			}
-
-			ploop_log(0, "Starting relocation");
-			ret = ploop_balloon_relocation(fd, &b_ctl, device);
-			break;
-		case PLOOP_DISCARD_STAT:
-			ploop_log(0, "Getting extents");
-			ret = discard_collect_stat(fd, distrib);
-			break;
-		default:
-			ret = -EINVAL;
+		/* serialize ploop operations vs ploop_complete_running_operation()
+		 * NB: PLOOP_IOC_BALLOON may change mntn from PLOOP_MNTN_DISCARD:
+		 * to PLOOP_MNTN_FBLOADED
+		 */
+		if (ploop_lock_di(di)) {
+			ret = SYSEXIT_LOCK;
 			break;
 		}
 
+		memset(&b_ctl, 0, sizeof(b_ctl));
+		b_ctl.keep_intact = 1;
+		ret = ioctl_device(fd, PLOOP_IOC_BALLOON, &b_ctl);
+		if (ret) {
+			ploop_unlock_di(di);
+			break;
+		}
+
+		if (b_ctl.mntn_type == PLOOP_MNTN_OFF) {
+			ploop_log(0, "Unexpected maintenance type 0x%x", b_ctl.mntn_type);
+			ret = -1;
+			ploop_unlock_di(di);
+			break;
+		}
+
+		if (size >= to_free || (stop && *stop)) {
+			ploop_log(3, "Killing the trim process %d", tpid);
+			kill(tpid, SIGUSR1);
+			ret = ioctl(fd, PLOOP_IOC_DISCARD_FINI);
+			if (ret < 0 && errno != EBUSY)
+				ploop_err(errno, "Can't finalize a discard mode");
+		}
+
+		ploop_log(0, "Starting relocation");
+		ret = ploop_balloon_relocation(fd, &b_ctl, device);
+		ploop_unlock_di(di);
 		if (ret)
 			break;
 	}
 
-	if (ret) {
-		exit_code = 1;
 
-		ret = ioctl_device(fd, PLOOP_IOC_DISCARD_FINI, NULL);
-		if (ret < 0)
-			ploop_err(errno, "Can't finalize discard mode");
+	if (ret) {
+		err = -1;
+
+		ret = ioctl(fd, PLOOP_IOC_DISCARD_FINI);
+		if (ret < 0) {
+			if (errno == EBUSY)
+				ploop_log(-1, "Discard finalized, but "
+					"relocation is still not completed");
+			else
+				ploop_err(errno, "Can't finalize discard mode");
+		}
 
 		kill(tpid, SIGKILL);
+	} else {
+		ploop_log(0, "%d clusters have been relocated", size);
 	}
 
 	ret = waitpid(tpid, &status, 0);
 	if (ret == -1) {
 		ploop_err(errno, "wait() failed");
-		exit_code = 1;
-	} else if(!WIFEXITED(status) || WEXITSTATUS(status)) {
-		if (WIFEXITED(status))
-			ploop_err(0, "The trim process failed with code %d",
-							WEXITSTATUS(status));
-		else
-			ploop_err(0, "The trim process killed by signal %d",
-							WTERMSIG(status));
-		exit_code = 1;
-	}
-
-	if (exit_code == 0 && state == PLOOP_DISCARD_STAT) {
-		int j;
-
-		for (j = DISCARD_TABLE_SIZE - 1; j >= 0; j--)
-			ploop_log(3, "%10d\t%u", j, distrib[j]);
-
-		for (j = DISCARD_TABLE_SIZE - 1; j >= 0; j--) {
-			if (to_free <= distrib[j]) {
-				*minlen_c = (1 << j);
-				break;
-			}
-
-			to_free -= distrib[j];
+		err = -1;
+	} else if (WIFEXITED(status)) {
+		ret = WEXITSTATUS(status);
+		if (ret) {
+			ploop_err(0, "The trim process failed with code %d", ret);
+			err = -1;
 		}
+	} else if (WIFSIGNALED(status)) {
+		ploop_err(0, "The trim process killed by signal %d", WTERMSIG(status));
+		err = -1;
+	} else {
+		ploop_err(0, "The trim process died abnormally");
+		err = -1;
 	}
 
-	return exit_code;
+	return err;
 }
 
-int ploop_discard(const char *device, const char *mount_point,
-				__u64 minlen_b, __u64 to_free)
+static int do_ploop_discard(struct ploop_disk_images_data *di,
+		const char *device, const char *mount_point,
+		__u64 minlen_b, __u64 to_free, const int *stop)
 {
-	int fd, ret, state;
+	int fd, ret;
 	int blocksize;
-	__u32  cluster, minlen_c;
+	__u32  cluster;
 
 	if (ploop_get_attr(device, "block_size", &blocksize)) {
 		ploop_err(0, "Can't find block size");
@@ -1080,30 +1023,211 @@ int ploop_discard(const char *device, const char *mount_point,
 	}
 	cluster = S2B(blocksize);
 
-	if (minlen_b || to_free == ~0ULL)
-		state = PLOOP_DISCARD_COMPACT;
-	else
-		state = PLOOP_DISCARD_STAT;
-
-	minlen_c = (minlen_b + cluster - 1) / cluster;
-	to_free = to_free  / cluster;
+	if (to_free == 0)
+		to_free = ~0ULL;
+	to_free = to_free / cluster;
 	if (!to_free) {
 		ploop_err(0, "Can't shrink by less than %d bytes", cluster);
-		return 0;
+		return -1;
 	}
 
 	fd = open_device(device);
 	if (fd == -1)
 		return SYSEXIT_OPEN;
 
-	for (; state < PLOOP_DISCARD_MAX; state++) {
-		ret = __ploop_discard(fd, device, mount_point, state,
-					&minlen_c, cluster, to_free);
-		if (ret)
-			break;
-	}
+	ret = __ploop_discard(di, fd, device, mount_point,
+					minlen_b, cluster, to_free, stop);
 
 	close(fd);
 
 	return ret;
 }
+
+static int ploop_get_dev_and_mnt(struct ploop_disk_images_data *di,
+		char *dev, int dev_len, char *mnt, int mnt_len)
+{
+	if (ploop_lock_di(di))
+		return SYSEXIT_LOCK;
+
+	if (ploop_find_dev(di->runtime->component_name,
+			di->images[0]->file, dev, dev_len))
+	{
+		ploop_unlock_di(di);
+		return SYSEXIT_PARAM;
+	}
+
+	if (ploop_get_mnt_by_dev(dev, mnt, mnt_len)) {
+		ploop_err(0, "Unable to find mount point for %s", dev);
+		ploop_unlock_di(di);
+		return SYSEXIT_PARAM;
+	}
+	ploop_unlock_di(di);
+
+	return 0;
+}
+
+int ploop_discard_by_dev(const char *device, const char *mount_point,
+		__u64 minlen_b, __u64 to_free, const int *stop)
+{
+	return do_ploop_discard(NULL, device, mount_point, minlen_b, to_free, stop);
+}
+
+int ploop_discard(struct ploop_disk_images_data *di,
+		struct ploop_discard_param *param)
+{
+	int ret;
+	char dev[PATH_MAX];
+	char mnt[PATH_MAX];
+	int mounted = 0;
+
+	if (ploop_lock_di(di))
+		return SYSEXIT_LOCK;
+
+	ret = ploop_find_dev(di->runtime->component_name,
+			di->images[0]->file, dev, sizeof(dev));
+	if (ret == -1) {
+		ploop_unlock_di(di);
+		return SYSEXIT_LOCK;
+	} else if (ret == 0) {
+		if (ploop_get_mnt_by_dev(dev, mnt, sizeof(mnt))) {
+			ploop_err(0, "Unable to find mount point for %s", dev);
+			ploop_unlock_di(di);
+			return SYSEXIT_PARAM;
+		}
+	} else {
+		struct ploop_mount_param mount_param = {};
+
+		if (!param->automount) {
+			ploop_err(0, "Unable to discard: image is not mounted");
+			ploop_unlock_di(di);
+			return SYSEXIT_PARAM;
+		}
+		ret = auto_mount_image(di, &mount_param);
+		if (ret) {
+			ploop_unlock_di(di);
+			return ret;
+		}
+		mounted = 1;
+		snprintf(dev, sizeof(dev), "%s", mount_param.device);
+		snprintf(mnt, sizeof(mnt), "%s", mount_param.target);
+
+		free_mount_param(&mount_param);
+	}
+	ploop_unlock_di(di);
+
+	ret = do_ploop_discard(di, dev, mnt, param->minlen_b,
+			param->to_free, param->stop);
+
+	if (mounted && ploop_lock_di(di) == 0) {
+		ploop_umount(dev, di);
+		ploop_unlock_di(di);
+	}
+
+	return ret;
+}
+
+int ploop_complete_running_operation(const char *device)
+{
+	struct ploop_balloon_ctl b_ctl;
+	int fd, ret;
+
+	fd = open_device(device);
+	if (fd == -1)
+		return SYSEXIT_OPEN;
+
+	bzero(&b_ctl, sizeof(b_ctl));
+	b_ctl.keep_intact = 1;
+	ret = ioctl(fd, PLOOP_IOC_BALLOON, &b_ctl);
+	if (ret) {
+		ploop_err(errno, "Unable to get in-kernel maintenance state");
+		ret = SYSEXIT_DEVIOC;
+		goto err;
+	}
+	if (b_ctl.mntn_type == PLOOP_MNTN_OFF)
+		goto err;
+
+	ploop_log(0, "Completing an on-going operation %s for device %s",
+			mntn2str(b_ctl.mntn_type), device);
+
+	switch (b_ctl.mntn_type) {
+		case PLOOP_MNTN_MERGE:
+			ret = ioctl_device(fd, PLOOP_IOC_MERGE, 0);
+			break;
+		case PLOOP_MNTN_GROW:
+			ret = ioctl_device(fd, PLOOP_IOC_GROW, 0);
+			break;
+		case PLOOP_MNTN_RELOC:
+		case PLOOP_MNTN_FBLOADED:
+			ret = ploop_balloon_complete(device);
+			break;
+		case PLOOP_MNTN_TRACK:
+			ret = ioctl_device(fd, PLOOP_IOC_TRACK_ABORT, 0);
+			break;
+		case PLOOP_MNTN_DISCARD:
+			ret = ploop_balloon_complete(device);
+			break;
+		case PLOOP_MNTN_BALLOON:
+			/*  FIXME : ploop_balloon_check_and_repair(device, mount_point, 1; */
+			ret = 0;
+			break;
+	}
+
+err:
+	close(fd);
+	return ret;
+}
+
+int ploop_discard_get_stat_by_dev(const char *device, const char *mount_point,
+		struct ploop_discard_stat *pd_stat)
+{
+	int		err;
+	struct statfs	stfs;
+	struct stat	st, balloon_stat;
+	off_t		ploop_size;
+	char		image[PATH_MAX];
+
+	err = get_balloon(mount_point, &balloon_stat, NULL);
+	if (err)
+		return err;
+
+	err = statfs(mount_point, &stfs);
+	if (err == -1) {
+		ploop_err(errno, "statfs(%s) failed", mount_point);
+		return 1;
+	}
+
+	err = ploop_get_size(device, &ploop_size);
+	if (err)
+		return 1;
+
+	err = ploop_find_top_delta_name_and_format(device, image, sizeof(image), NULL, 0);
+	if (err)
+		return 1;
+
+	err = stat(image, &st);
+	if (err == -1) {
+		ploop_err(errno, "stat(%s) failed", image);
+		return 1;
+	}
+
+	pd_stat->ploop_size = S2B(ploop_size) - balloon_stat.st_size;
+	pd_stat->image_size = st.st_size;
+	pd_stat->data_size = pd_stat->ploop_size - stfs.f_bfree * stfs.f_bsize;
+	pd_stat->balloon_size = balloon_stat.st_size;
+
+	return 0;
+}
+
+int ploop_discard_get_stat(struct ploop_disk_images_data *di,
+		struct ploop_discard_stat *pd_stat)
+{
+	int ret;
+	char dev[PATH_MAX];
+	char mnt[PATH_MAX];
+
+	ret = ploop_get_dev_and_mnt(di, dev, sizeof(dev), mnt, sizeof(mnt));
+	if (ret)
+		return ret;
+
+	return ploop_discard_get_stat_by_dev(dev, mnt, pd_stat);
+}
