@@ -24,10 +24,11 @@
 #include <malloc.h>
 #include <string.h>
 #include <dirent.h>
+#include <limits.h>
 #include <sys/stat.h>
-#include <sys/param.h>
 #include <linux/fs.h>
 #include <sys/ioctl.h>
+#include <sys/utsname.h>
 
 #include "ploop.h"
 
@@ -48,21 +49,18 @@ int ploop_find_top_delta_name_and_format(
 	char *img;
 	char *fmt;
 
-	if (ploop_get_attr(device, "top", &top_level)) {
-		ploop_err(0, "Can't find top delta");
+	if (ploop_get_attr(device, "top", &top_level))
+		return SYSEXIT_SYSFS;
+
+	if (find_delta_names(device, top_level, top_level, &img, &fmt))
 		return SYSEXIT_SYSFS;
-	}
 
-	if (find_delta_names(device, top_level, top_level,
-			     &img, &fmt)) {
-		ploop_err(errno, "find_delta_names");
-		return(SYSEXIT_SYSFS);
-	}
 	if (image)
 		strncpy(image, img, image_size);
 	free(img);
 	if (format)
 		strncpy(format, fmt, format_size);
+
 	return 0;
 }
 
@@ -70,62 +68,34 @@ int find_delta_names(const char * device, int start_level, int end_level,
 		     char **names, char ** format)
 {
 	int i;
-	FILE * fp;
-	int len;
+	char path[PATH_MAX];
 	char nbuf[4096];
 
 	if (memcmp(device, "/dev/", 5) == 0)
 		device += 5;
 
 	for (i = 0; i <= end_level - start_level; i++) {
-		snprintf(nbuf, sizeof(nbuf)-1, "/sys/block/%s/pdelta/%d/image",
+		snprintf(path, sizeof(path), "/sys/block/%s/pdelta/%d/image",
 			 device, start_level + i);
 
-		fp = fopen(nbuf, "r");
-		if (fp == NULL) {
-			ploop_err(errno, "fopen sysfs image %s", nbuf);
+		if (read_line(path, nbuf, sizeof(nbuf)))
 			return -1;
-		}
-		if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-			ploop_err(errno, "read sysfs image");
-			fclose(fp);
-			return -1;
-		}
-		len = strlen(nbuf);
-		if (len > 0 && nbuf[len-1] == '\n') {
-			len--;
-			nbuf[len] = 0;
-		}
-		names[(end_level-start_level)-i] = strdup(nbuf);
-		fclose(fp);
 
-		if (i == 0) {
-			snprintf(nbuf, sizeof(nbuf)-1, "/sys/block/%s/pdelta/%d/format",
+		names[(end_level-start_level)-i] = strdup(nbuf);
+
+		if (i == 0 && format) {
+			snprintf(path, sizeof(path), "/sys/block/%s/pdelta/%d/format",
 				 device, start_level);
-			fp = fopen(nbuf, "r");
-			if (fp == NULL) {
-				ploop_err(errno, "fopen sysfs format %s", nbuf);
+
+			if (read_line(path, nbuf, sizeof(nbuf)))
 				return -1;
-			}
-			if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-				ploop_err(errno, "read sysfs format");
-				fclose(fp);
-				return -1;
-			}
-			len = strlen(nbuf);
-			if (len > 0 && nbuf[len-1] == '\n') {
-				len--;
-				nbuf[len] = 0;
-			}
-			if (format) {
-				if (strcmp(nbuf, "raw") == 0)
-					*format = "raw";
-				else if (strcmp(nbuf, "ploop1") == 0)
-					*format = "ploop1";
-				else
-					*format = "unknown";
-			}
-			fclose(fp);
+
+			if (strcmp(nbuf, "raw") == 0)
+				*format = "raw";
+			else if (strcmp(nbuf, "ploop1") == 0)
+				*format = "ploop1";
+			else
+				*format = "unknown";
 		}
 	}
 	return 0;
@@ -133,75 +103,43 @@ int find_delta_names(const char * device, int start_level, int end_level,
 
 int ploop_get_attr(const char * device, const char * attr, int * res)
 {
-	FILE * fp;
+	char path[PATH_MAX];
 	char nbuf[4096];
 
 	if (memcmp(device, "/dev/", 5) == 0)
 		device += 5;
 
-	snprintf(nbuf, sizeof(nbuf)-1, "/sys/block/%s/pstate/%s", device, attr);
+	snprintf(path, sizeof(path), "/sys/block/%s/pstate/%s", device, attr);
 
-	fp = fopen(nbuf, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen %s", nbuf);
+	if (read_line(path, nbuf, sizeof(nbuf)))
 		return -1;
-	}
-
-	if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-		ploop_err(errno, "fgets");
-		fclose(fp);
-		return -1;
-	}
-	fclose(fp);
 
 	if (sscanf(nbuf, "%d", res) != 1) {
-		ploop_err(0, "Unexpected format of %s/pstate/%s %s",
-			device, attr, nbuf);
+		ploop_err(0, "Unexpected format of %s: %s", path, nbuf);
 		return -1;
 	}
 	return 0;
 }
 
-static int ploop_get_delta_attr_str(const char *device, int level, char * attr, char *nbuf, int nbuf_len)
+int ploop_get_delta_attr(const char *device, int level, const char *attr, int *res)
 {
-	FILE * fp;
+	char path[PATH_MAX];
+	char nbuf[4096];
 
 	if (memcmp(device, "/dev/", 5) == 0)
 		device += 5;
 
-	snprintf(nbuf, nbuf_len-1, "/sys/block/%s/pdelta/%d/%s",
+	snprintf(path, sizeof(path), "/sys/block/%s/pdelta/%d/%s",
 			device,	level, attr);
 
-	fp = fopen(nbuf, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen %s", nbuf);
+	if (read_line(path, nbuf, sizeof(nbuf)))
 		return -1;
-	}
-
-	if (fgets(nbuf, nbuf_len, fp) == NULL) {
-		ploop_err(errno, "fgets /sys/block/%s/pdelta/%d/%s",
-				device, level, attr);
-		fclose(fp);
-		return -1;
-	}
-	fclose(fp);
-
-	return 0;
-}
-
-int ploop_get_delta_attr(const char * device, int level, char * attr, int * res)
-{
-	char nbuf[4096];
-	int err;
-
-	if ((err = ploop_get_delta_attr_str(device, level, attr, nbuf, sizeof(nbuf))))
-		return err;
 
 	if (sscanf(nbuf, "%d", res) != 1) {
-		ploop_err(0, "Unexpected format of %s/pdelta/%s %s",
-			device, attr, nbuf);
+		ploop_err(0, "Unexpected format of %s: %s", path, nbuf);
 		return -1;
 	}
+
 	return 0;
 }
 
@@ -214,8 +152,7 @@ int ploop_get_size(const char * device, off_t * res)
 		ploop_err(errno, "Can't open %s", device);
 		return SYSEXIT_OPEN;
 	}
-	if (ioctl(fd, BLKGETSIZE64, res) < 0) {
-		ploop_err(errno, "ioctl(BLKGETSIZE)");
+	if (ioctl_device(fd, BLKGETSIZE64, res)) {
 		close(fd);
 		return SYSEXIT_BLKDEV;
 	}
@@ -225,71 +162,33 @@ int ploop_get_size(const char * device, off_t * res)
 	return 0;
 }
 
-int read_line(const char *path, char *nbuf, int len)
-{
-	FILE *fp;
-
-	fp = fopen(path, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen %s", path);
-		return -1;
-	}
-	if (fgets(nbuf, len, fp) == NULL) {
-		ploop_err(errno, "read %s", path);
-		fclose(fp);
-		return -1;
-	}
-	fclose(fp);
-	len = strlen(nbuf);
-	if (len > 0 && nbuf[len-1] == '\n') {
-		nbuf[len-1] = 0;
-	}
-	return 0;
-}
-
-static int get_dev_num(char *path, dev_t *dev_num)
+static int get_dev_num(const char *path, dev_t *dev_num)
 {
 	char nbuf[4096];
 	int maj, min;
 
 	if (read_line(path, nbuf, sizeof(nbuf)))
-		return-1;
+		return -1;
+
 	if (sscanf(nbuf, "%d:%d", &maj, &min) != 2) {
-		ploop_err(0, "Unexpected format of /sys/.../dev: %s", nbuf);
+		ploop_err(0, "Unexpected format of %s: %s", path, nbuf);
 		return -1;
 	}
 	*dev_num = gnu_dev_makedev(maj, min);
 	return 0;
 }
 
-static int get_dev_start(char *path, __u32 *start)
+static int get_dev_start(const char *path, __u32 *start)
 {
-	FILE * fp;
-	int len;
 	char nbuf[4096];
 
-	fp = fopen(path, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen %s", path);
+	if (read_line(path, nbuf, sizeof(nbuf)))
 		return -1;
-	}
-	if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-		ploop_err(errno, "read sysfs start");
-		fclose(fp);
-		return -1;
-	}
-	len = strlen(nbuf);
-	if (len > 0 && nbuf[len-1] == '\n') {
-		len--;
-		nbuf[len] = 0;
-	}
 
 	if (sscanf(nbuf, "%u", start) != 1) {
-		ploop_err(0, "Unexpected format of /sys/.../start: %s", nbuf);
-		fclose(fp);
+		ploop_err(0, "Unexpected format of %s: %s", path, nbuf);
 		return -1;
 	}
-	fclose(fp);
 
 	return 0;
 }
@@ -316,7 +215,7 @@ int dev_num2dev_start(const char *device, dev_t dev_num, __u32 *dev_start)
 	snprintf(nbuf, sizeof(nbuf) - 1, "/sys/block/%s", device);
 	dp = opendir(nbuf);
 	if (dp == NULL) {
-		ploop_err(errno, "sysfs opendir");
+		ploop_err(errno, "Can't opendir %s", nbuf);
 		return -1;
 	}
 
@@ -331,7 +230,7 @@ int dev_num2dev_start(const char *device, dev_t dev_num, __u32 *dev_start)
 		snprintf(nbuf, sizeof(nbuf) - 1, "/sys/block/%s/%s",
 			 device, de->d_name);
 		if (lstat(nbuf, &st)) {
-			ploop_err(errno, "lstat");
+			ploop_err(errno, "Can't lstat %s", nbuf);
 			goto close_dir;
 		}
 
@@ -359,75 +258,115 @@ close_dir:
 	return -1;
 }
 
-int ploop_find_dev_by_delta(char *delta, char *buf, int size)
+static int is_cookie_supported(void)
 {
-	char fname[MAXPATHLEN];
-	char image[MAXPATHLEN];
+	struct utsname buf;
+	int major, minor;
+
+	uname(&buf);
+	// >= 2.6.32-042stab061.1
+	if (sscanf(buf.release, "%*d.%*d.%*d-%dstab%d.%*d", &major, &minor) != 2) {
+		ploop_err(0, "Can't parse kernel version %s", buf.release);
+		return 0;
+	}
+
+	return (major > 42 || (major == 42 && minor >= 61));
+}
+
+/* Find device by base delta and return the name
+ * return: -1 on error
+ *	    0 found
+ *	    1 not found
+ */
+int ploop_find_dev(const char *component_name, const char *delta,
+		char *buf, int size)
+{
+	char fname[PATH_MAX];
+	char delta_r[PATH_MAX];
+	char image[PATH_MAX];
 	DIR *dp;
 	struct dirent *de;
-	struct stat st, st1;
-	int ret = 1;
-	char name[64];
-	dev_t dev;
+	int ret = -1;
+	char cookie[PLOOP_COOKIE_SIZE];
+	int lckfd;
 
-	if (stat(delta, &st)) {
-		if (errno == ENOENT)
-			return 1;
-		ploop_err(errno, "ploop_find_dev_by_delta stat(%s)",
-				delta);
+	if (realpath(delta, delta_r) == NULL) {
+		ploop_err(errno, "Can't resolve %s", delta);
 		return -1;
 	}
 
+	lckfd = ploop_global_lock();
+	if (lckfd == -1)
+		return -1;
+
 	snprintf(fname, sizeof(fname) - 1, "/sys/block/");
 	dp = opendir(fname);
 	if (dp == NULL) {
-		ploop_err(errno, "opendir %s", fname);
-		return -1;
+		ploop_err(errno, "Can't opendir %s", fname);
+		goto err;
 	}
 
 	while ((de = readdir(dp)) != NULL) {
-		if (strncmp("ploop", de->d_name, 4))
+		int err;
+
+		if (strncmp("ploop", de->d_name, 5))
 			continue;
 
 		snprintf(fname, sizeof(fname), "/sys/block/%s/pdelta/0/image",
 				de->d_name);
-		if (stat(fname, &st1))
-			continue;
-		if (read_line(fname, image, sizeof(image)))
-			continue;
-		if (stat(image, &st1))
-			continue;
-		if (st.st_dev != st1.st_dev || st.st_ino != st1.st_ino)
+		err = read_line_quiet(fname, image, sizeof(image));
+		if (err) {
+			if (err == ENOENT)
+				continue;
+
+			ploop_err(err, "Can't open or read %s", fname);
+			goto err;
+		}
+
+		if (strcmp(image, delta_r))
 			continue;
 
-		snprintf(fname, sizeof(fname), "/sys/block/%s/dev",
+		snprintf(fname, sizeof(fname), "/sys/block/%s/pstate/cookie",
 				de->d_name);
-		if (get_dev_num(fname, &dev) == 0) {
-			snprintf(buf, size, "/dev/%s",
-					make_sysfs_dev_name(gnu_dev_minor(dev), name, sizeof(name)));
-			if (stat(buf, &st1) == 0 &&
-					st1.st_rdev != dev)
-			{
-				ploop_err(0, "Inconsistency in device number detected for %s sys_dev=%lu dev=%lu",
-						buf, (unsigned long)dev, (unsigned long)st1.st_rdev);
-				ret = -1;
-			} else
-				ret = 0;
-			break;
+		err = read_line_quiet(fname, cookie, sizeof(cookie));
+		if (err) {
+			if ((err == ENOENT) && is_cookie_supported())
+				/* This is not an error, but a race between
+				 * mount and umount: device is being removed
+				 */
+				continue;
+
+			ploop_err(err, "Can't open or read %s", fname);
+			if ((errno == ENOENT) && component_name)
+				/* Using component_name on old kernel is bad */
+				ploop_err(0, "ERROR: OpenVZ kernel with ploop cookie support "
+						"(i.e. 042stab061.1 or greater) is required");
+			goto err;
 		}
+
+		if (strncmp(component_name == NULL ? "" : component_name,
+					cookie, sizeof(cookie)))
+			continue;
+
+		snprintf(buf, size, "/dev/%s", de->d_name);
+		ret = 0;
+		goto err;
 	}
-	closedir(dp);
+	ret = 1; /* not found */
+
+err:
+	if (dp)
+		closedir(dp);
+	close(lckfd);
 
 	return ret;
 }
 
 int ploop_get_top_level(int devfd, const char *devname, int *top)
 {
-	char path[MAXPATHLEN];
+	char path[PATH_MAX];
 	char name[64];
 	struct stat st;
-	FILE * fp;
-	int len;
 	char nbuf[4096];
 
 	if (fstat(devfd, &st)) {
@@ -438,30 +377,15 @@ int ploop_get_top_level(int devfd, const char *devname, int *top)
 	snprintf(path, sizeof(path) - 1, "/sys/block/%s/pstate/top",
 		 make_sysfs_dev_name(gnu_dev_minor(st.st_rdev), name, sizeof(name)));
 
-	fp = fopen(path, "r");
-	if (fp == NULL) {
-		ploop_err(errno, "fopen %s (%s)", path, devname);
+	if (read_line(path, nbuf, sizeof(nbuf)))
 		return -1;
-	}
-	if (fgets(nbuf, sizeof(nbuf), fp) == NULL) {
-		ploop_err(errno, "fgets from %s (%s)", path, devname);
-		fclose(fp);
-		return -1;
-	}
-	len = strlen(nbuf);
-	if (len > 0 && nbuf[len-1] == '\n') {
-		len--;
-		nbuf[len] = 0;
-	}
 
 	if (sscanf(nbuf, "%d", top) != 1) {
 		ploop_err(0, "Unexpected format of %s: %s (%s)",
 			  path, nbuf, devname);
-		fclose(fp);
 		errno = ERANGE;
 		return -1;
 	}
-	fclose(fp);
 
 	return 0;
 }
