@@ -24,7 +24,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
-#include <sys/param.h>
+#include <limits.h>
 #include <sys/ioctl.h>
 #include <sys/mount.h>
 #include <linux/blkpg.h>
@@ -122,7 +122,6 @@ int get_partition_device_name(const char *device, char *out, int size)
 
 static int blkpg_resize_partition(int fd, struct GptEntry *pe)
 {
-	int ret;
 	struct blkpg_ioctl_arg ioctl_arg;
 	struct blkpg_partition part;
 
@@ -138,16 +137,9 @@ static int blkpg_resize_partition(int fd, struct GptEntry *pe)
 	ioctl_arg.datalen = sizeof(struct blkpg_partition);
 	ioctl_arg.data = &part;
 
-	ret = ioctl(fd, BLKPG, &ioctl_arg);
-	if (ret)
-		ploop_err(errno, "BLKPG_RESIZE_PARTITION");
-
-	return ret;
+	return ioctl_device(fd, BLKPG, &ioctl_arg);
 }
 
-#define GPT_HEADER_SIZE		1
-#define	GPT_PT_ENTRY_SIZE	32
-#define GPT_DATA_SIZE		(GPT_HEADER_SIZE + GPT_PT_ENTRY_SIZE)
 int resize_gpt_partition(const char *devname)
 {
 	unsigned char buf[SECTOR_SIZE*GPT_DATA_SIZE]; // LBA1 header, LBA2-34 partition entry
@@ -190,9 +182,9 @@ int resize_gpt_partition(const char *devname)
 		goto err;
 	}
 	// change GPT header
-	pt->alternate_lba = size -1;
-	pt->last_usable_lba = size - GPT_DATA_SIZE -1;
-	pe->ending_lba = pt->last_usable_lba;
+	pt->alternate_lba = size - 1;
+	pt->last_usable_lba = size - GPT_DATA_SIZE - 1;
+	pe->ending_lba = (pt->last_usable_lba >> 3 << 3) - 1;
 
 	// Recalculate crc32
 	pe_crc32 = crc32((unsigned char *)pe, SECTOR_SIZE * GPT_PT_ENTRY_SIZE);
