@@ -24,7 +24,7 @@
 #include <malloc.h>
 #include <sys/stat.h>
 #include <sys/mount.h>
-#include <sys/param.h>
+#include <limits.h>
 #include <getopt.h>
 #include <linux/types.h>
 #include <string.h>
@@ -188,13 +188,10 @@ static int grow_lower_delta(const char *device, int top, int start_level, int en
 
 	/* update in-core map_node mappings for relocated blocks */
 	grow_maps.ctl->level = start_level;
-	if (ioctl(devfd, PLOOP_IOC_UPDATE_INDEX, grow_maps.ctl) < 0) {
-		ploop_err(errno, "PLOOP_IOC_UPDATE_INDEX");
-		ret = SYSEXIT_DEVIOC;
-		close(devfd);
-		goto done;
-	}
+	ret = ioctl_device(devfd, PLOOP_IOC_UPDATE_INDEX, grow_maps.ctl);
 	close(devfd);
+	if (ret)
+		goto done;
 
 	/* nullify relocated blocks on disk */
 	memset(buf, 0, cluster);
@@ -242,9 +239,8 @@ int get_delta_info(const char *device, struct merge_info *info)
 		return SYSEXIT_PARAM;
 	}
 
-	if (info->end_level == 0) {
+	if (info->end_level == 0)
 		info->end_level = info->top_level;
-	}
 
 	if (info->end_level > info->top_level ||
 	    info->start_level > info->end_level)
@@ -306,6 +302,10 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 	}
 
 	if (device) {
+		ret = ploop_complete_running_operation(device);
+		if (ret)
+			return ret;
+
 		if (merge_top) {
 			/* top delta is in running state merged
 			by means of PLOOP_IOC_MERGE */
@@ -329,14 +329,10 @@ int merge_image(const char *device, int start_level, int end_level, int raw, int
 				return SYSEXIT_DEVICE;
 			}
 
-			if (ioctl(lfd, PLOOP_IOC_MERGE, 0) < 0) {
-				ploop_err(errno, "PLOOP_IOC_MERGE");
-				close(lfd);
-				return SYSEXIT_DEVIOC;
-			}
-
+			ret = ioctl_device(lfd, PLOOP_IOC_MERGE, 0);
 			close(lfd);
-			return 0;
+
+			return ret;
 		}
 		last_delta = end_level - start_level;
 	} else {
@@ -580,9 +576,8 @@ merge_done:
 		level = start_level + 1;
 
 		for (i = start_level + 1; i <= end_level; i++) {
-			if (ioctl(lfd, PLOOP_IOC_DEL_DELTA, &level) < 0) {
-				ploop_err(errno, "PLOOP_IOC_DEL_DELTA");
-				ret = SYSEXIT_DEVIOC;
+			ret = ioctl_device(lfd, PLOOP_IOC_DEL_DELTA, &level);
+			if (ret) {
 				close(lfd);
 				goto merge_done2;
 			}
@@ -590,10 +585,7 @@ merge_done:
 
 		if (merge_top) {
 			ploop_log(0, "Merging top delta");
-			if (ioctl(lfd, PLOOP_IOC_MERGE, 0) < 0) {
-				ploop_err(errno, "PLOOP_IOC_MERGE");
-				ret = SYSEXIT_DEVIOC;
-			}
+			ret = ioctl_device(lfd, PLOOP_IOC_MERGE, 0);
 		}
 
 		close(lfd);
@@ -606,7 +598,7 @@ merge_done2:
 
 int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode)
 {
-	char conf[MAXPATHLEN];
+	char conf[PATH_MAX];
 	char dev[64];
 	char *device = NULL;
 	char *fname = NULL;
@@ -776,7 +768,7 @@ err:
 
 int ploop_merge_snapshot(struct ploop_disk_images_data *di, struct ploop_merge_param *param)
 {
-	int ret;
+	int ret = SYSEXIT_PARAM;
 	const char *guid = NULL;
 
 	if (ploop_lock_di(di))
