@@ -27,7 +27,7 @@
 #include <sys/stat.h>
 #include <sys/mount.h>
 #include <sys/sysmacros.h>
-#include <sys/param.h>
+#include <limits.h>
 #include <sys/file.h>
 #include <fcntl.h>
 #include <sys/vfs.h>
@@ -35,25 +35,10 @@
 #include <mntent.h>
 
 #include "ploop.h"
-
-static __thread struct ploop_cancel_handle __cancel_data;
+#include "cleanup.h"
 
 static int ploop_mount_fs(struct ploop_mount_param *param);
 
-struct ploop_cancel_handle *ploop_get_cancel_handle(void)
-{
-	return &__cancel_data;
-}
-
-static int not_supported_for_vm(struct ploop_disk_images_data *di)
-{
-	if (!di->runtime->vm_compat)
-		return 0;
-
-	ploop_err(0, "Operation is not supported for VM");
-	return SYSEXIT_PARAM;
-}
-
 /* set cancel flag
  * Note: this function also clear the flag
  */
@@ -69,13 +54,7 @@ static int is_operation_cancelled(void)
 	return 0;
 }
 
-void ploop_cancel_operation(struct ploop_cancel_handle *handle)
-{
-	ploop_log(0, "cancel operation");
-	handle->flags = 1;
-}
-
-static void free_mount_param(struct ploop_mount_param *param)
+void free_mount_param(struct ploop_mount_param *param)
 {
 	free(param->target);
 	free(param->guid);
@@ -83,7 +62,7 @@ static void free_mount_param(struct ploop_mount_param *param)
 
 static off_t bytes2sec(__u64 bytes)
 {
-	return (bytes >> 9) + ((bytes % 512) ? 1 : 0);
+	return (bytes >> PLOOP1_SECTOR_LOG) + ((bytes % SECTOR_SIZE) ? 1 : 0);
 }
 
 int sys_fallocate(int fd, int mode, off_t offset, off_t len)
@@ -104,7 +83,7 @@ int get_list_size(char **list)
 	return i;
 }
 
-static char **make_images_list(struct ploop_disk_images_data *di, char *guid, int reverse)
+static char **make_images_list(struct ploop_disk_images_data *di, const char *guid, int reverse)
 {
 	int n;
 	char **images;
@@ -256,6 +235,26 @@ static int get_temp_mountpoint(const char *file, int create, char *buf, int len)
 	return 0;
 }
 
+static int check_blockdev_size(unsigned long long sectors, __u32 blocksize)
+{
+	const unsigned long long max = (__u32)-1;
+
+	if (sectors > max) {
+		ploop_err(0, "An incorrect block device size is specified: %llu sectors."
+				" The maximum allowed size is %llu sectors.",
+				sectors, max);
+		return -1;
+	}
+	if (sectors % blocksize) {
+		ploop_err(0, "An incorrect block device size is specified: %llu sectors."
+				" The block device size must be aligned to the cluster block size %d.",
+				sectors, blocksize);
+		return -1;
+	}
+
+	return 0;
+}
+
 static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize)
 {
 	int fd;
@@ -266,7 +265,7 @@ static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize)
 
 	assert(blocksize);
 
-	if (bdsize > (__u32)-1)
+	if (check_blockdev_size(bdsize, blocksize))
 		return -1;
 
 	if (posix_memalign(&buf, 4096, cluster)) {
@@ -276,7 +275,7 @@ static int create_empty_delta(const char *path, __u32 blocksize, off_t bdsize)
 
 	ploop_log(0, "Creating delta %s bs=%d size=%ld sectors",
 			path, blocksize, (long)bdsize);
-	fd = open(path, O_RDWR|O_CREAT|O_EXCL, 0600);
+	fd = open(path, O_RDWR|O_CREAT|O_DIRECT|O_EXCL, 0600);
 	if (fd < 0) {
 		ploop_err(errno, "Can't open %s", path);
 		free(buf);
@@ -326,7 +325,7 @@ static int create_empty_preallocated_delta(const char *path, __u32 blocksize, of
 	off_t off;
 	__u64 cluster = S2B(blocksize);
 
-	if (bdsize > (__u32)-1)
+	if (check_blockdev_size(bdsize, blocksize))
 		return -1;
 
 	if (posix_memalign(&buf, 4096, cluster)) {
@@ -349,10 +348,14 @@ static int create_empty_preallocated_delta(const char *path, __u32 blocksize, of
 
 	rc = sys_fallocate(odelta.fd, 0, 0, S2B(vh.m_FirstBlockOffset + vh.m_SizeInSectors));
 	if (rc) {
-		if (errno == ENOTSUP)
-			ploop_err(errno, "fallocate");
-		ploop_err(errno, "Failed to create %s", path);
-		goto out_close;
+		if (errno == ENOTSUP) {
+			ploop_log(0, "Warning: fallocate is not supported, using truncate instead");
+			rc = ftruncate(odelta.fd, S2B(vh.m_FirstBlockOffset + vh.m_SizeInSectors));
+		}
+		if (rc) {
+			ploop_err(errno, "Failed to create %s", path);
+			goto out_close;
+		}
 	}
 
 	for (clu = 0; clu < SizeToFill / cluster; clu++) {
@@ -361,7 +364,7 @@ static int create_empty_preallocated_delta(const char *path, __u32 blocksize, of
 
 		if (clu > 0)
 			memset(buf, 0, cluster);
-		for (i = (clu == 0 ? 16 : 0); i < (cluster / sizeof(__u32)) &&
+		for (i = (clu == 0 ? PLOOP_MAP_OFFSET : 0); i < (cluster / sizeof(__u32)) &&
 				l2_slot < vh.m_Size;
 				i++, l2_slot++)
 		{
@@ -373,7 +376,7 @@ static int create_empty_preallocated_delta(const char *path, __u32 blocksize, of
 	}
 
 	if (fsync(odelta.fd)) {
-		ploop_err(errno, "fsync");
+		ploop_err(errno, "fsync %s", path);
 		goto out_close;
 	}
 	free(buf);
@@ -468,7 +471,7 @@ static int create_image(struct ploop_disk_images_data *di,
 	int fd = -1;
 	int ret;
 	struct ploop_pvd_header vh = {};
-	char fname[MAXPATHLEN];
+	char fname[PATH_MAX];
 	struct stat st;
 
 	if (size_sec == 0 || file == NULL)
@@ -501,7 +504,7 @@ static int create_image(struct ploop_disk_images_data *di,
 		goto err;
 	}
 
-	if (ploop_di_add_image(di, fname, BASE_UUID, NONE_UUID)) {
+	if (ploop_di_add_image(di, fname, TOPDELTA_UUID, NONE_UUID)) {
 		ret = SYSEXIT_NOMEM;
 		goto err;
 	}
@@ -518,11 +521,11 @@ err:
 }
 
 static int create_balloon_file(struct ploop_disk_images_data *di,
-		char *device, char *fstype)
+		const char *device)
 {
 	int fd, ret;
-	char mnt[MAXPATHLEN];
-	char fname[MAXPATHLEN + sizeof(BALLOON_FNAME)];
+	char mnt[PATH_MAX];
+	char fname[PATH_MAX + sizeof(BALLOON_FNAME)];
 	struct ploop_mount_param mount_param = {};
 
 	if (device == NULL)
@@ -582,7 +585,7 @@ static int ploop_init_image(struct ploop_disk_images_data *di, struct ploop_crea
 	ret = make_fs(mount_param.device, param->fstype);
 	if (ret)
 		goto err;
-	ret = create_balloon_file(di, mount_param.device, param->fstype);
+	ret = create_balloon_file(di, mount_param.device);
 	if (ret)
 		goto err;
 
@@ -597,7 +600,7 @@ err:
 static int ploop_drop_image(struct ploop_disk_images_data *di)
 {
 	int i;
-	char fname[MAXPATHLEN];
+	char fname[PATH_MAX];
 
 	if (di->nimages == 0)
 		return SYSEXIT_PARAM;
@@ -633,12 +636,13 @@ int ploop_create_image(struct ploop_create_param *param)
 		return SYSEXIT_PARAM;
 	}
 
-	di = ploop_alloc_diskdescriptor();
+	di = alloc_diskdescriptor();
 	if (di == NULL)
 		return SYSEXIT_NOMEM;
 	di->blocksize = blocksize;
 	ret = create_image(di, param->image, di->blocksize,
-			param->size, param->mode);
+			ROUNDUP_BDSIZE(param->size, di->blocksize),
+			param->mode);
 	if (ret)
 		return ret;
 	if (param->fstype != NULL) {
@@ -651,34 +655,31 @@ int ploop_create_image(struct ploop_create_param *param)
 	return ret;
 }
 
+#define PROC_PLOOP_MINOR	"/proc/vz/ploop_minor"
+
 int ploop_getdevice(int *minor)
 {
-	int lfd;
-	struct stat st;
-	struct ploop_getdevice_ctl req;
-	int ret = 0;
-	const char *dev = "/dev/ploop0";
+	int fd, ret;
+	char buf[64];
 
-	if (stat(dev, &st))
-		if (mknod(dev, S_IFBLK, gnu_dev_makedev(PLOOP_DEV_MAJOR, 0))) {
-			ploop_err(errno, "mknod %s", dev);
-			return SYSEXIT_MKNOD;
-		}
-
-	lfd = open(dev, O_RDONLY);
-	if (lfd < 0) {
-		ploop_err(errno, "Can't open device %s", dev);
-		return SYSEXIT_DEVICE;
+	fd = open(PROC_PLOOP_MINOR, O_RDONLY);
+	if (fd < 0) {
+		ploop_err(errno, "Can't open " PROC_PLOOP_MINOR);
+		return -1;
+	}
+	ret = read(fd, buf, sizeof(buf));
+	if (ret == -1) {
+		ploop_err(errno, "Can't read from " PROC_PLOOP_MINOR);
+		close(fd);
+		return -1;
+	}
+	if (sscanf(buf, "%d", minor) != 1) {
+		ploop_err(0, "Can't get ploop minor '%s'", buf);
+		close(fd);
+		return -1;
 	}
 
-	if (ioctl(lfd, PLOOP_IOC_GETDEVICE, &req) < 0) {
-		ploop_err(errno, "PLOOP_IOC_GETDDEVICE");
-		ret = SYSEXIT_DEVIOC;
-	}
-	*minor = req.minor;
-
-	close(lfd);
-	return ret;
+	return fd;
 }
 
 /* Workaround for bug #PCLIN-30116 */
@@ -695,18 +696,46 @@ static int do_ioctl(int fd, int req)
 	return ret;
 }
 
+static int print_lsof(int level, const char * message, const char * mnt, int err)
+{
+	FILE * lsoff;
+	char buffer[8192]; /* same as LOG_BUF_SIZE */
+
+	snprintf(buffer, sizeof(buffer), "/usr/sbin/lsof %s", mnt);
+	if ((lsoff = popen(buffer, "r")) == NULL) {
+		ploop_err(errno, "Can't exec lsof");
+		if (level == -1)
+			ploop_err(err, "%s %s", message, mnt);
+		else
+			ploop_log(level, "%s %s", message, mnt);
+
+		return -1;
+	}
+
+	while (fgets(buffer, sizeof(buffer), lsoff))
+		ploop_log(level, "%s %s: %s, lsof output follows:\n%s",
+				message, mnt, strerror(err), buffer);
+	pclose(lsoff);
+
+	return 0;
+}
+
 static int do_umount(const char *mnt)
 {
 	int i, ret;
 
-	for (i = 0; i < 60; i++) {
+	for (i = 0; i < 6; i++) {
 		ret = umount(mnt);
 		if (ret == 0 || (ret == -1 && errno != EBUSY))
 			return ret;
-		ploop_log(3, "retry umount %s", mnt);
+		if (ploop_get_log_level() > 3)
+			print_lsof(3, "Retrying umount", mnt, errno);
+		else
+			ploop_log(3, "Retrying umount %s", mnt);
 		sleep(1);
 	}
-	return ret;
+	print_lsof(-1, "Can't umount", mnt, errno);
+	return -1;
 }
 
 static int delete_deltas(int devfd, const char *devname)
@@ -749,10 +778,11 @@ static int ploop_stop(int fd, const char *devname)
 static int get_mount_dir(const char *dev, char *buf, int size)
 {
 	FILE *fp;
-	struct mntent *ent;
+	struct mntent *ent, mntbuf;
 	int ret = 1;
 	int len;
 	const char *ep;
+	char tmp[512];
 
 	len = strlen(dev);
 	if (len == 0)
@@ -763,7 +793,7 @@ static int get_mount_dir(const char *dev, char *buf, int size)
 		ploop_err(errno, "Can't open /proc/mounts");
 		return -1;
 	}
-	while ((ent = getmntent(fp))) {
+	while ((ent = getmntent_r(fp, &mntbuf, tmp, sizeof(tmp)))) {
 		ep = ent->mnt_fsname + len;
 		// check for /dev/ploopN or /dev/ploopNp1
 		if (strncmp(dev, ent->mnt_fsname, len) == 0 &&
@@ -921,7 +951,7 @@ int ploop_get_dev(struct ploop_disk_images_data *di, char *out, int len)
 	return ret;
 }
 
-static int reread_part(char *device)
+static int reread_part(const char *device)
 {
 	int fd;
 
@@ -942,7 +972,7 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 	unsigned long flags =
 		(param->flags & MS_NOATIME) |
 		(param->ro ? MS_RDONLY : 0);
-	char buf[MAXPATHLEN + sizeof(BALLOON_FNAME)];
+	char buf[PATH_MAX + sizeof(BALLOON_FNAME)];
 	struct stat st;
 	char *fstype = param->fstype == NULL ? DEFAULT_FSTYPE : param->fstype;
 	char data[1024];
@@ -987,30 +1017,51 @@ static int ploop_mount_fs(struct ploop_mount_param *param)
 	return 0;
 }
 
-static int add_delta(int lfd, char *image, struct ploop_ctl_delta *req)
+static int add_delta(int lfd, const char *component_name,
+		const char *image, struct ploop_ctl_delta *req, int level)
 {
 	int fd;
-	int ro = (req->c.pctl_flags == PLOOP_FMT_RDONLY);
+	int ro = (req->c.pctl_flags & PLOOP_FMT_RDONLY);
+	int ret;
 
-	fd = open(image, ro ? O_RDONLY : O_RDWR);
+	fd = open(image, O_DIRECT | (ro ? O_RDONLY : O_RDWR));
 	if (fd < 0) {
 		ploop_err(errno, "Can't open file %s", image);
-		close(fd);
 		return SYSEXIT_OPEN;
 	}
 
 	req->f.pctl_fd = fd;
 
+	if (level == 0) {
+		/* Base image: check if in use */
+		char dev[64];
+		int rc;
+
+		ret = SYSEXIT_PARAM;
+		rc = ploop_find_dev(component_name, image, dev, sizeof(dev));
+		if (rc < 0)
+			goto out;
+		if (rc == 0) {
+			ret = SYSEXIT_MOUNT;
+			ploop_err(0, "Image %s already used by device %s",
+					image, dev);
+			goto out;
+		}
+	}
+
 	if (ioctl(lfd, PLOOP_IOC_ADD_DELTA, req) < 0) {
 		ploop_err(0, "Can't add image %s: %s", image,
 				(errno == ENOTSUP) ?
 					"unsupported underlying filesystem"
 					: strerror(errno));
-		close(fd);
-		return SYSEXIT_DEVIOC;
+		ret = SYSEXIT_DEVIOC;
+		goto out;
 	}
+	ret = 0;
+out:
 	close(fd);
-	return 0;
+
+	return ret;
 }
 
 static int create_ploop_dev(int minor)
@@ -1049,77 +1100,71 @@ static int add_deltas(struct ploop_disk_images_data *di,
 	char *device = param->device;
 	int i;
 	int ret = 0;
-	int registered = 0;
+	struct ploop_ctl_delta req = {};
+	const char *component_name = di ? di->runtime->component_name : NULL;
 
 	if (device[0] == '\0') {
 		char buf[64];
 		int minor;
 
-		lckfd = ploop_global_lock();
+		lckfd = ploop_getdevice(&minor);
 		if (lckfd == -1)
-			return SYSEXIT_LOCK;
+			return SYSEXIT_DEVICE;
 
-		ret = ploop_getdevice(&minor);
-		if (ret)
-			goto err;
 		snprintf(device, sizeof(param->device), "/dev/%s",
 				make_sysfs_dev_name(minor, buf, sizeof(buf)));
 		ret = create_ploop_dev(minor);
 		if (ret)
 			goto err;
-		if (di != NULL) {
-			ret = register_ploop_dev(di->runtime->component_name,
-					images[0], device);
-			if (ret)
-				goto err;
-			registered = 1;
-		}
-
 	}
 
 	*lfd_p = open(device, O_RDONLY);
 	if (*lfd_p < 0) {
 		ploop_err(errno, "Can't open device %s", device);
 		ret = SYSEXIT_DEVICE;
-		goto err1;
+		goto err;
 	}
 
+	if (di != NULL && di->runtime->component_name != NULL) {
+		req.c.pctl_flags |= PLOOP_FLAG_COOKIE;
+		strncpy(req.cookie, di->runtime->component_name,
+				PLOOP_COOKIE_SIZE);
+	}
+	req.c.pctl_cluster_log = ffs(blocksize) - 1;
+	req.c.pctl_size = 0;
+	req.c.pctl_chunks = 1;
+
+	req.f.pctl_fd = -1;
+	req.f.pctl_type = PLOOP_IO_AUTO;
+
 	for (i = 0; images[i] != NULL; i++) {
-		struct ploop_ctl_delta req;
 		int ro = (images[i+1] != NULL || param->ro) ? 1: 0;
 		char *image = images[i];
 
-		memset(&req, 0, sizeof(req));
-
 		req.c.pctl_format = PLOOP_FMT_PLOOP1;
 		if (raw && i == 0)
 			req.c.pctl_format = PLOOP_FMT_RAW;
 		if (ro)
-			req.c.pctl_flags = PLOOP_FMT_RDONLY;
-		req.c.pctl_cluster_log = ffs(blocksize) - 1;
-		req.c.pctl_size = 0;
-		req.c.pctl_chunks = 1;
-
-		req.f.pctl_fd = -1;
-		req.f.pctl_type = PLOOP_IO_DIRECT;
-		if (ploop_is_on_nfs(image))
-			req.f.pctl_type = PLOOP_IO_NFS;
+			req.c.pctl_flags |= PLOOP_FMT_RDONLY;
+		else
+			req.c.pctl_flags &= ~PLOOP_FMT_RDONLY;
 
 		ploop_log(0, "Adding delta dev=%s img=%s (%s)",
 				device, image, ro ? "ro" : "rw");
-		ret = add_delta(*lfd_p, image, &req);
+		ret = add_delta(*lfd_p, component_name, image, &req, i);
 		if (ret)
 			goto err1;
 	}
 	if (ioctl(*lfd_p, PLOOP_IOC_START, 0) < 0) {
 		ploop_err(errno, "PLOOP_IOC_START");
 		ret = SYSEXIT_DEVIOC;
-		goto err;
+		goto err1;
 	}
 
 err1:
-	if (ret && *lfd_p != -1) {
+	if (ret) {
 		int err = 0;
+		int empty = !i;
 
 		for (i = i - 1; i >= 0; i--) {
 			err = ioctl(*lfd_p, PLOOP_IOC_DEL_DELTA, &i);
@@ -1128,13 +1173,12 @@ err1:
 				break;
 			}
 		}
-		if (err == 0 && ioctl(*lfd_p, PLOOP_IOC_CLEAR, 0) < 0)
+		if (!empty && err == 0 && ioctl(*lfd_p, PLOOP_IOC_CLEAR, 0) < 0)
 			ploop_err(errno, "PLOOP_IOC_CLEAR");
 	}
-	if (ret && registered)
-		unregister_ploop_dev(di->runtime->component_name, images[0]);
 err:
-	ploop_unlock(&lckfd);
+	if (lckfd != -1)
+		close(lckfd);
 	return ret;
 }
 
@@ -1152,9 +1196,15 @@ int ploop_mount(struct ploop_disk_images_data *di, char **images,
 		return SYSEXIT_PARAM;
 	}
 
-	if (param->target != NULL && stat(param->target, &st)) {
-		ploop_err(0, "Mount point %s does not exist", param->target);
-		return SYSEXIT_PARAM;
+	if (param->target != NULL) {
+		if (stat(param->target, &st)) {
+			ploop_err(0, "Mount point %s does not exist", param->target);
+			return SYSEXIT_PARAM;
+		}
+		if (!S_ISDIR(st.st_mode)) {
+			ploop_err(0, "Mount point %s not a directory", param->target);
+			return SYSEXIT_PARAM;
+		}
 	}
 
 	if (raw) {
@@ -1209,6 +1259,11 @@ err:
 	if (lfd >= 0)
 		close(lfd);
 
+	if (ret == 0 && di != NULL &&
+			di->runtime->component_name == NULL &&
+			param->target != NULL)
+		drop_statfs_info(di->images[0]->file);
+
 	return ret;
 }
 
@@ -1246,43 +1301,18 @@ static int mount_image(struct ploop_disk_images_data *di, struct ploop_mount_par
 	return ret;
 }
 
-static int auto_mount_image(struct ploop_disk_images_data *di, struct ploop_mount_param *param, int *mounted)
+int auto_mount_image(struct ploop_disk_images_data *di,
+		struct ploop_mount_param *param)
 {
-	char mnt[MAXPATHLEN];
-	int ret = 0;
-	int dev_mounted = 0;
-	int fs_mounted = 0;
+	char mnt[PATH_MAX];
+	int ret;
 
-	ret = ploop_find_dev_by_uuid(di, 1, param->device, sizeof(param->device));
-	if (ret == -1)
-		return SYSEXIT_MOUNT;
-	if (ret == 0) {
-		dev_mounted = 1;
-		ret = get_mount_dir(param->device, mnt, sizeof(mnt));
-		if (ret == -1) {
-			ploop_err(0, "Can't find mount point for %s",
-				param->device);
-			return SYSEXIT_MOUNT;
-		} else if (ret == 0) {
-			param->target = strdup(mnt);
-			fs_mounted = 1;
-		}
-	}
-	if (!fs_mounted) {
-		ret = get_temp_mountpoint(di->images[0]->file, 1, mnt, sizeof(mnt));
-		if (ret)
-			return ret;
-		param->target = strdup(mnt);
-	}
+	ret = get_temp_mountpoint(di->images[0]->file, 1, mnt, sizeof(mnt));
+	if (ret)
+		return ret;
+	param->target = strdup(mnt);
 
-	if (!dev_mounted) {
-		ret = mount_image(di, param, 0);
-		if (ret == 0)
-			*mounted = 1;
-	} else if (!fs_mounted) {
-		ret = ploop_mount_fs(param);
-	}
-	return ret;
+	return mount_image(di, param, 0);
 }
 
 int ploop_mount_image(struct ploop_disk_images_data *di, struct ploop_mount_param *param)
@@ -1299,7 +1329,7 @@ int ploop_mount_image(struct ploop_disk_images_data *di, struct ploop_mount_para
 		return -1;
 	}
 	if (ret == 0) {
-		ploop_err(0, "Image %s already mounted to %s",
+		ploop_err(0, "Image %s already used by device %s",
 				di->images[0]->file, dev);
 
 		ret = SYSEXIT_MOUNT;
@@ -1315,7 +1345,11 @@ err:
 
 int ploop_mount_snapshot(struct ploop_disk_images_data *di, struct ploop_mount_param *param)
 {
-	return mount_image(di, param, PLOOP_MOUNT_SNAPSHOT);
+	if (param->guid == NULL) {
+		ploop_err(0, "Snapshot guid is not specified");
+		return SYSEXIT_PARAM;
+	}
+	return ploop_mount_image(di, param);
 }
 
 static int ploop_stop_device(const char *device)
@@ -1338,8 +1372,7 @@ static int ploop_stop_device(const char *device)
 int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 {
 	int ret;
-	char mnt[MAXPATHLEN] = "";
-	int lckfd;
+	char mnt[PATH_MAX] = "";
 
 	if (!device) {
 		ploop_err(0, "ploop_umount: device is not specified");
@@ -1347,7 +1380,10 @@ int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 	}
 
 	if (get_mount_dir(device, mnt, sizeof(mnt)) == 0) {
-		if (di != NULL)
+		/* The component_name feature allows multiple image mount.
+		 * Skip store statfs in custom case.
+		 */
+		if (di != NULL && di->runtime->component_name == NULL)
 			store_statfs_info(mnt, di->images[0]->file);
 		ploop_log(0, "Unmounting file system at %s", mnt);
 		if (do_umount(mnt)) {
@@ -1356,24 +1392,15 @@ int ploop_umount(const char *device, struct ploop_disk_images_data *di)
 		}
 	}
 
-	lckfd = ploop_global_lock();
-	if (lckfd == -1)
-		return SYSEXIT_LOCK;
-
 	ret = ploop_stop_device(device);
 
-	if (ret == 0 && di != NULL)
-		unregister_ploop_dev(di->runtime->component_name, di->images[0]->file);
-
-	ploop_unlock(&lckfd);
-
 	return ret;
 }
 
 int ploop_umount_image(struct ploop_disk_images_data *di)
 {
 	int ret;
-	char dev[MAXPATHLEN];
+	char dev[PATH_MAX];
 
 	if (di->nimages == 0) {
 		ploop_err(0, "No images specified");
@@ -1394,6 +1421,12 @@ int ploop_umount_image(struct ploop_disk_images_data *di)
 		return -1;
 	}
 
+	ret = ploop_complete_running_operation(dev);
+	if (ret) {
+		ploop_unlock_di(di);
+		return ret;
+	}
+
 	ret = ploop_umount(dev, di);
 
 	ploop_unlock_di(di);
@@ -1447,19 +1480,24 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 {
 	int ret;
 	struct ploop_mount_param mount_param = {};
-	char buf[MAXPATHLEN];
-	int were_mounted = 0;
+	char buf[PATH_MAX];
+	int mounted = 0;
 	int balloonfd = -1;
 	struct stat st;
 	off_t dev_size = 0;
 	__u64 balloon_size = 0;
 	__u64 new_balloon_size = 0;
 	struct statfs fs;
+	unsigned long long new_size = ROUNDUP_BDSIZE(param->size, di->blocksize);
 
 	if (di->nimages == 0) {
 		ploop_err(0, "No images in DiskDescriptor");
 		return -1;
 	}
+
+	if (check_blockdev_size(new_size, di->blocksize))
+		return -1;
+
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
@@ -1467,10 +1505,15 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 	if (ret == -1)
 		goto err;
 	if (ret != 0) {
-		ret = auto_mount_image(di, &mount_param, &were_mounted);
+		ret = auto_mount_image(di, &mount_param);
 		if (ret)
 			goto err;
+		mounted = 1;
 	} else {
+		ret = ploop_complete_running_operation(buf);
+		if (ret)
+			goto err;
+
 		strncpy(mount_param.device, buf, sizeof(mount_param.device));
 		if (get_mount_dir(mount_param.device, buf, sizeof(buf))) {
 			ploop_err(0, "Can't find mount point for %s", buf);
@@ -1507,9 +1550,9 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 		new_balloon_size -= B2S(delta);
 		ret = ploop_balloon_change_size(mount_param.device,
 				balloonfd, new_balloon_size);
-	} else if (param->size > dev_size) {
-		char conf[MAXPATHLEN];
-		char conf_tmp[MAXPATHLEN];
+	} else if (new_size > dev_size) {
+		char conf[PATH_MAX];
+		char conf_tmp[PATH_MAX];
 
 		// GROW
 		if (balloon_size != 0) {
@@ -1519,14 +1562,14 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 				goto err;
 		}
 		// Update size in the DiskDescriptor.xml
-		di->size = param->size;
+		di->size = new_size;
 		get_disk_descriptor_fname(di, conf, sizeof(conf));
 		snprintf(conf_tmp, sizeof(conf_tmp), "%s.tmp", conf);
 		ret = ploop_store_diskdescriptor(conf_tmp, di);
 		if (ret)
 			goto err;
 
-		ret = ploop_grow_device(mount_param.device, di->blocksize, param->size);
+		ret = ploop_grow_device(mount_param.device, di->blocksize, new_size);
 		if (ret) {
 			unlink(conf_tmp);
 			goto err;
@@ -1542,7 +1585,7 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 		ret = resize_fs(mount_param.device);
 		if (ret)
 			goto err;
-		tune_fs(mount_param.target, mount_param.device, param->size);
+		tune_fs(mount_param.target, mount_param.device, new_size);
 	} else {
 		off_t available_balloon_size;
 		// SHRINK
@@ -1553,12 +1596,12 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 			ret = SYSEXIT_FSTAT;
 			goto err;
 		}
-		new_balloon_size = dev_size - param->size;
+		new_balloon_size = dev_size - new_size;
 		available_balloon_size = balloon_size + B2S(fs.f_bfree * fs.f_bsize);
 		if (available_balloon_size < new_balloon_size) {
 			ploop_err(0, "Unable to change image size to %llu "
 					"sectors, minimal size is %lu",
-					param->size,
+					new_size,
 					(long unsigned)(dev_size - available_balloon_size));
 			ret = SYSEXIT_PARAM;
 			goto err;
@@ -1569,13 +1612,13 @@ int ploop_resize_image(struct ploop_disk_images_data *di, struct ploop_resize_pa
 					balloonfd, new_balloon_size);
 			if (ret)
 				goto err;
-			tune_fs(mount_param.target, mount_param.device, param->size);
+			tune_fs(mount_param.target, mount_param.device, new_size);
 		}
 	}
 
 err:
 	close(balloonfd);
-	if (were_mounted)
+	if (mounted)
 		ploop_umount(mount_param.device, di);
 	ploop_unlock_di(di);
 	free_mount_param(&mount_param);
@@ -1589,8 +1632,7 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 	struct delta odelta = {};
 	__u32 clu;
 	void *buf = NULL;
-	char tmp[MAXPATHLEN];
-	char fname[MAXPATHLEN];
+	char tmp[PATH_MAX];
 	int ret = -1;
 	__u64 cluster;
 
@@ -1607,8 +1649,7 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 
 	snprintf(tmp, sizeof(tmp), "%s.tmp",
 			di->images[0]->file);
-	ret = open_delta_simple(&odelta, tmp, O_RDWR|O_CREAT|O_EXCL|O_TRUNC, OD_OFFLINE);
-	if (ret)
+	if (open_delta_simple(&odelta, tmp, O_RDWR|O_CREAT|O_EXCL|O_TRUNC, OD_OFFLINE))
 		goto err;
 
 	for (clu = 0; clu < delta.l2_size; clu++) {
@@ -1645,10 +1686,6 @@ static int expanded2raw(struct ploop_disk_images_data *di)
 	if (fsync(odelta.fd))
 		ploop_err(errno, "fsync");
 
-	get_disk_descriptor_fname(di, fname, sizeof(fname));
-	if (ploop_store_diskdescriptor(fname, di))
-		goto err;
-
 	if (rename(tmp, di->images[0]->file)) {
 		ploop_err(errno, "rename %s %s",
 			tmp, di->images[0]->file);
@@ -1672,17 +1709,16 @@ static int expanded2preallocated(struct ploop_disk_images_data *di)
 	off_t data_off;
 	int ret = -1;
 	__u64 cluster;
+	void *buf = NULL;
 
 	ploop_log(0, "Converting image to preallocated...");
 	// FIXME: deny on snapshots
 	if (open_delta(&delta, di->images[0]->file, O_RDWR, OD_OFFLINE))
 		return SYSEXIT_OPEN;
+
 	cluster = S2B(delta.blocksize);
 	data_off = delta.alloc_head;
-	if (fsync(delta.fd)) {
-		ploop_err(errno, "fsync");
-		goto err;
-	}
+
 	// Second stage: update index
 	for (clu = 0; clu < delta.l2_size; clu++) {
 		int l2_cluster = (clu + PLOOP_MAP_OFFSET) / (cluster / sizeof(__u32));
@@ -1700,15 +1736,28 @@ static int expanded2preallocated(struct ploop_disk_images_data *di)
 		}
 		if (delta.l2[l2_slot] == 0) {
 			off_t idx_off = (off_t)l2_cluster * cluster + (l2_slot*sizeof(__u32));
+			int rc;
 
 			delta.l2[l2_slot] = data_off * delta.blocksize;
 
-			ret = sys_fallocate(delta.fd, 0, data_off * cluster, cluster);
-			if (ret) {
-				if (errno == ENOTSUP)
-					ploop_err(errno, "fallocate");
-				ploop_err(errno, "Failed to expand %s", di->images[0]->file);
-				goto err;
+			rc = sys_fallocate(delta.fd, 0, data_off * cluster, cluster);
+			if (rc) {
+				if (errno == ENOTSUP) {
+					if (buf == NULL) {
+						ploop_log(0, "Warning: fallocate is not supported,"
+								" using write instead");
+						buf = calloc(1, cluster);
+						if (buf == NULL) {
+							ploop_err(errno, "malloc");
+							goto err;
+						}
+					}
+					rc = PWRITE(&delta, buf, cluster, data_off * cluster);
+				}
+				if (rc) {
+					ploop_err(errno, "Failed to expand %s", di->images[0]->file);
+					goto err;
+				}
 			}
 
 			if (PWRITE(&delta, &delta.l2[l2_slot], sizeof(__u32), idx_off))
@@ -1724,13 +1773,14 @@ static int expanded2preallocated(struct ploop_disk_images_data *di)
 	ret = 0;
 err:
 	close_delta(&delta);
+	free(buf);
 	return ret;
 }
 
 int ploop_convert_image(struct ploop_disk_images_data *di, int mode, int flags)
 {
-	char conf_tmp[MAXPATHLEN];
-	char conf[MAXPATHLEN];
+	char conf_tmp[PATH_MAX];
+	char conf[PATH_MAX];
 	int ret = -1;
 
 	if (di->mode == PLOOP_RAW_MODE) {
@@ -1775,23 +1825,43 @@ err:
 	return ret;
 }
 
-int ploop_get_info(struct ploop_disk_images_data *di, struct ploop_info *info)
+static int ploop_get_info(struct ploop_disk_images_data *di, struct ploop_info *info)
 {
-	char mnt[MAXPATHLEN];
+	char mnt[PATH_MAX];
 	char dev[64];
+	int ret = -1;
 
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 	if (ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev)) == 0 &&
 			get_mount_dir(dev, mnt, sizeof(mnt)) == 0)
 	{
-		ploop_unlock_di(di);
-		if (get_statfs_info(mnt, info) == 0)
-			return 0;
+		ret = get_statfs_info(mnt, info);
 	}
 
 	ploop_unlock_di(di);
-	return read_statfs_info(di->images[0]->file, info);
+
+	return ret;
+}
+
+int ploop_get_info_by_descr(const char *descr, struct ploop_info *info)
+{
+	struct ploop_disk_images_data *di;
+	int ret;
+
+	/* Try the fast path first, for stopped ploop */
+	if (read_statfs_info(descr, info) == 0)
+		return 0;
+
+	ret = ploop_read_disk_descr(&di, descr);
+	if (ret)
+		return ret;
+
+	ret = ploop_get_info(di, info);
+
+	ploop_free_diskdescriptor(di);
+
+	return ret;
 }
 
 static int do_snapshot(int lfd, int fd, struct ploop_ctl_delta *req)
@@ -1814,6 +1884,10 @@ int create_snapshot(const char *device, const char *delta, __u32 blocksize, int
 	__u64 bdsize;
 	struct ploop_ctl_delta req;
 
+	ret = ploop_complete_running_operation(device);
+	if (ret)
+		return ret;
+
 	lfd = open(device, O_RDONLY);
 	if (lfd < 0) {
 		ploop_err(errno, "Can't open device %s", device);
@@ -1847,9 +1921,7 @@ int create_snapshot(const char *device, const char *delta, __u32 blocksize, int
 	req.c.pctl_size = 0;
 	req.c.pctl_chunks = 1;
 
-	req.f.pctl_type = PLOOP_IO_DIRECT;
-	if (ploop_is_on_nfs(delta))
-		req.f.pctl_type = PLOOP_IO_NFS;
+	req.f.pctl_type = PLOOP_IO_AUTO;
 
 	ploop_log(0, "Creating snapshot dev=%s img=%s", device, delta);
 	ret = do_snapshot(lfd, fd, &req);
@@ -1913,19 +1985,14 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 	int ret;
 	int fd;
 	char dev[64];
-	char uuid[61];
-	char uuid1[61];
-	char fname[MAXPATHLEN];
-	char conf[MAXPATHLEN];
-	char conf_tmp[MAXPATHLEN];
+	char snap_guid[61];
+	char file_guid[61];
+	char fname[PATH_MAX];
+	char conf[PATH_MAX];
+	char conf_tmp[PATH_MAX];
 	int online = 0;
 	int n;
 	off_t size;
-	char *top_guid = NULL;
-
-	ret = not_supported_for_vm(di);
-	if (ret)
-		return ret;
 
 	if (di->nimages == 0) {
 		ploop_err(0, "No images");
@@ -1939,11 +2006,13 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
-	ret = gen_uuid_pair(uuid, sizeof(uuid), uuid1, sizeof(uuid1));
+	ret = gen_uuid_pair(snap_guid, sizeof(snap_guid),
+			file_guid, sizeof(file_guid));
 	if (ret) {
 		ploop_err(errno, "Can't generate uuid");
 		goto err_cleanup1;
 	}
+
 	if (param->guid != NULL) {
 		if (find_snapshot_by_guid(di, param->guid) != -1) {
 			ploop_err(0, "The snapshot %s already exist",
@@ -1951,10 +2020,7 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 			ret = SYSEXIT_PARAM;
 			goto err_cleanup1;
 		}
-		if (param->snap_guid)
-			ploop_di_change_guid(di, di->top_guid, param->guid);
-		else
-			strcpy(uuid, param->guid);
+		strcpy(snap_guid, param->guid);
 	}
 	n = get_snapshot_count(di);
 	if (n == -1) {
@@ -1971,10 +2037,10 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 		goto err_cleanup1;
 	}
 
-	top_guid = strdup(di->top_guid);
 	snprintf(fname, sizeof(fname), "%s.%s",
-			di->images[0]->file, uuid1);
-	ret = ploop_di_add_image(di, fname, uuid, get_top_delta_guid(di));
+			di->images[0]->file, file_guid);
+	ploop_di_change_guid(di, di->top_guid, snap_guid);
+	ret = ploop_di_add_image(di, fname, TOPDELTA_UUID, snap_guid);
 	if (ret)
 		goto err_cleanup1;
 
@@ -1990,7 +2056,7 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 
 	if (ret != 0) {
 		// offline snapshot
-		ret = get_image_size(di, top_guid, &size);
+		ret = get_image_size(di, snap_guid, &size);
 		if (ret)
 			goto err_cleanup2;
 		fd = create_empty_delta(fname, di->blocksize, size);
@@ -2019,14 +2085,13 @@ int ploop_create_snapshot(struct ploop_disk_images_data *di, struct ploop_snapsh
 				fname);
 
 	ploop_log(0, "ploop snapshot %s has been successfully created",
-			di->top_guid);
+			snap_guid);
 err_cleanup2:
 	if (ret && !online && unlink(conf_tmp))
-		ploop_err(errno, "Can't unlink %s",
-				conf_tmp);
+		ploop_err(errno, "Can't unlink %s", conf_tmp);
+
 err_cleanup1:
 	ploop_unlock_di(di);
-	free(top_guid);
 
 	return ret;
 }
@@ -2037,16 +2102,17 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 	int fd;
 	char dev[64];
 	char uuid[61];
-	char uuid1[61];
-	char new_top_delta_fname[MAXPATHLEN];
+	char file_uuid[61];
+	char new_top_delta_fname[PATH_MAX];
 	char *old_top_delta_fname = NULL;
-	char conf[MAXPATHLEN];
-	char conf_tmp[MAXPATHLEN];
+	char conf[PATH_MAX];
+	char conf_tmp[PATH_MAX];
 	off_t size;
 
-	ret = not_supported_for_vm(di);
-	if (ret)
-		return ret;
+	if (guid == NULL || !is_valid_guid(guid)) {
+		ploop_err(0, "Incorrect guid %s", guid);
+		return SYSEXIT_PARAM;
+	}
 
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
@@ -2068,12 +2134,17 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 	if (ret)
 		goto err_cleanup1;
 
-	ret = gen_uuid_pair(uuid, sizeof(uuid), uuid1, sizeof(uuid1));
+	ret = gen_uuid_pair(uuid, sizeof(uuid), file_uuid, sizeof(file_uuid));
 	if (ret) {
 		ploop_err(errno, "Can't generate uuid");
 		goto err_cleanup1;
 	}
-	if (!(flags & PLOOP_LEAVE_TOP_DELTA)) {
+
+	ret = ploop_di_remove_image(di, di->top_guid, 0, &old_top_delta_fname);
+	if (ret)
+		goto err_cleanup1;
+
+	if (!(flags & PLOOP_SNAP_SKIP_TOPDELTA_DESTROY)) {
 		// device should be stopped
 		ret = ploop_find_dev_by_uuid(di, 1, dev, sizeof(dev));
 		if (ret == -1) {
@@ -2081,18 +2152,17 @@ int ploop_switch_snapshot(struct ploop_disk_images_data *di, const char *guid, i
 			goto err_cleanup1;
 		} else if (ret == 0) {
 			ret = SYSEXIT_PARAM;
-			ploop_err(0, "Unable to perform switch to snapshot operation on running device (%s)",
+			ploop_err(0, "Unable to perform switch to snapshot operation"
+					" on running device (%s)",
 					dev);
 			goto err_cleanup1;
 		}
+	} else
+		old_top_delta_fname = NULL;
 
-		ret = ploop_di_remove_image(di, di->top_guid, &old_top_delta_fname);
-		if (ret)
-			goto err_cleanup1;
-	}
 	snprintf(new_top_delta_fname, sizeof(new_top_delta_fname), "%s.%s",
-			di->images[0]->file, uuid1);
-	ret = ploop_di_add_image(di, new_top_delta_fname, uuid, guid);
+			di->images[0]->file, file_uuid);
+	ret = ploop_di_add_image(di, new_top_delta_fname, TOPDELTA_UUID, guid);
 	if (ret)
 		goto err_cleanup1;
 
@@ -2151,16 +2221,12 @@ int ploop_delete_top_delta(struct ploop_disk_images_data *di)
 int ploop_delete_snapshot(struct ploop_disk_images_data *di, const char *guid)
 {
 	int ret;
-	char conf[MAXPATHLEN];
+	char conf[PATH_MAX];
 	char *fname = NULL;
 	int nelem = 0;
 	char dev[64];
 	int snap_id;
 
-	ret = not_supported_for_vm(di);
-	if (ret)
-		return ret;
-
 	if (ploop_lock_di(di))
 		return SYSEXIT_LOCK;
 
@@ -2189,7 +2255,7 @@ int ploop_delete_snapshot(struct ploop_disk_images_data *di, const char *guid)
 			goto err;
 		}
 		/* snapshot is not active and last -> delete */
-		ret = ploop_di_remove_image(di, guid, &fname);
+		ret = ploop_di_remove_image(di, guid, 1, &fname);
 		if (ret)
 			goto err;
 		get_disk_descriptor_fname(di, conf, sizeof(conf));
