@@ -32,34 +32,32 @@
 #endif
 #endif /* ! __NR_syncfs */
 
-#define DISKDESCRIPTOR_XML      "DiskDescriptor.xml"
+#define DISKDESCRIPTOR_XML	"DiskDescriptor.xml"
 /* Compatibility: Parallels use this UUID to mark top delta */
-#define BASE_UUID		"{5fbaabe3-6958-40ff-92a7-860e329aab41}"
+#define TOPDELTA_UUID		"{5fbaabe3-6958-40ff-92a7-860e329aab41}"
 #define NONE_UUID		"{00000000-0000-0000-0000-000000000000}"
 #define DEFAULT_FSTYPE		"ext4"
 #define BALLOON_FNAME		".balloon-c3a5ae3d-ce7f-43c4-a1ea-c61e2b4504e8"
 
 /* od_flags for open_delta() */
-#define OD_NOFLAGS      0x0
-#define OD_ALLOW_DIRTY  0x1
-#define OD_OFFLINE      0x2
+#define OD_NOFLAGS	0x0
+#define OD_ALLOW_DIRTY	0x1
+#define OD_OFFLINE	0x2
 
 /* fsck mode flags */
-#define FSCK_FORCE     0x01
-#define FSCK_HARDFORCE 0x02
-#define FSCK_DROPINUSE 0x04
-#define FSCK_CHECK     0x08
+#define FSCK_FORCE	0x01
+#define FSCK_HARDFORCE	0x02
+#define FSCK_DROPINUSE	0x04
+#define FSCK_CHECK	0x08
+
+/* GPT */
+#define GPT_HEADER_SIZE		1
+#define GPT_PT_ENTRY_SIZE	32
+#define GPT_DATA_SIZE		(GPT_HEADER_SIZE + GPT_PT_ENTRY_SIZE)
 
 #define S2B(sec) ((off_t)(sec) << PLOOP1_SECTOR_LOG)
 #define B2S(sec) ((sec) >> PLOOP1_SECTOR_LOG)
-
-/* mount flags */
-#define PLOOP_MOUNT_SNAPSHOT   0x01
-
-struct ploop_cancel_handle
-{
-	int flags;
-};
+#define ROUNDUP_BDSIZE(bdsize, blocksize) ((bdsize + blocksize - 1) / blocksize * blocksize)
 
 struct delta_fops
 {
@@ -161,7 +159,6 @@ struct xfer_desc
 struct ploop_disk_images_runtime_data {
 	int lckfd;
 	char *xml_fname;
-	int vm_compat;
 	char *component_name;
 };
 
@@ -178,7 +175,7 @@ int gen_uuid_pair(char *uuid1, int len1, char *uuid2, int len2);
 int find_delta_names(const char * device, int start_level, int end_level,
 			    char **names, char ** format);
 PL_EXT int ploop_get_attr(const char * device, const char * attr, int * res);
-int ploop_get_delta_attr(const char * device, int level, char * attr, int * res);
+int ploop_get_delta_attr(const char * device, int level, const char * attr, int * res);
 int ploop_get_size(const char * device, off_t * res);
 int dev_num2dev_start(const char *device, dev_t dev_num, __u32 *dev_start);
 int ploop_get_top_level(int devfd, const char *devname, int *top);
@@ -208,9 +205,9 @@ int rmap2freemap(__u32 *rmap, __u32 iblk_start, __u32 iblk_end,
 struct ploop_freeblks_ctl;
 int freeblks_alloc(struct ploop_freeblks_ctl **freeblks_pp, int n);
 int freemap2freeblks(struct freemap *freemap,
-	       int lvl, struct ploop_freeblks_ctl **freeblks_pp, __u32 *total);
+		int lvl, struct ploop_freeblks_ctl **freeblks_pp, __u32 *total);
 int freeblks2freemap(struct ploop_freeblks_ctl *freeblks,
-		       struct freemap **freemap_pp, __u32 *total);
+		struct freemap **freemap_pp, __u32 *total);
 
 int range_build(__u32 a_h, __u32 n_free_blocks,
 		__u32 *rmap, __u32 rlen,
@@ -225,6 +222,7 @@ int relocmap2relocblks(struct relocmap *relocmap, int lvl, __u32 a_h, __u32 n_sc
 			struct ploop_relocblks_ctl **relocblks_pp);
 PL_EXT int ploop_fsck(char *img, int flags, int ro, int verbose, __u32 *blocksize_p);
 /* Logging */
+int ploop_get_log_level(void);
 void ploop_log(int level, const char *format, ...)
 	__attribute__ ((__format__ (__printf__, 2, 3)));
 void __ploop_err(int err_no, const char *format, ...)
@@ -239,6 +237,16 @@ void __ploop_err(int err_no, const char *format, ...)
 #define ploop_err __ploop_err
 #endif
 
+#define ioctl_device(fd, req, arg)					\
+	({								\
+		int __ret = 0;						\
+		if (ioctl(fd, req, arg)) {				\
+			ploop_err(errno, "Error in ioctl(" #req ")");	\
+			__ret = SYSEXIT_DEVIOC;				\
+		}							\
+		__ret;							\
+	 })
+
 char *make_sysfs_dev_name(int minor, char *buf, int len);
 PL_EXT int ploop_mount(struct ploop_disk_images_data *di, char **images,
 		struct ploop_mount_param *param, int raw);
@@ -247,6 +255,8 @@ int get_list_size(char **list);
 void free_images_list(char **images);
 int PWRITE(struct delta * delta, void * buf, unsigned int size, off_t off);
 int PREAD(struct delta * delta, void *buf, unsigned int size, off_t off);
+PL_EXT int ploop_getdevice(int *minor);
+struct ploop_disk_images_data *alloc_diskdescriptor(void);
 void get_disk_descriptor_fname(struct ploop_disk_images_data *di, char *buf, int size);
 void get_disk_descriptor_lock_fname(struct ploop_disk_images_data *di, char *out, int size);
 int ploop_find_dev_by_uuid(struct ploop_disk_images_data *di, int check_state, char *out, int len);
@@ -256,7 +266,8 @@ int sys_syncfs(int fd);
 // manage struct ploop_disk_images_data
 int ploop_di_add_image(struct ploop_disk_images_data *di, const char *fname,
 		const char *guid, const char *parent_guid);
-int ploop_di_remove_image(struct ploop_disk_images_data *di, const char *guid, char **fname);
+int ploop_di_remove_image(struct ploop_disk_images_data *di, const char *guid,
+		int renew_top_uuid, char **fname);
 int ploop_di_merge_image(struct ploop_disk_images_data *di, const char *guid, char **fname);
 void ploop_di_change_guid(struct ploop_disk_images_data *di, const char *guid, const char *new_guid);
 char *find_image_by_guid(struct ploop_disk_images_data *di, const char *guid);
@@ -265,8 +276,6 @@ int ploop_add_image_entry(struct ploop_disk_images_data *di, const char *fname,
 int ploop_add_snapshot_entry(struct ploop_disk_images_data *di, const char *guid,
 		const char *parent_guid);
 int ploop_find_dev(const char *module, const char *image, char *out, int size);
-int register_ploop_dev(const char *module, const char *image, const char *dev);
-void unregister_ploop_dev(const char *module, const char *image);
 
 //balloon
 PL_EXT char *mntn2str(int mntn_type);
@@ -275,9 +284,11 @@ PL_EXT int ploop_balloon_change_size(const char *device, int balloonfd, off_t ne
 PL_EXT int ploop_balloon_get_state(const char *device, __u32 *state);
 PL_EXT int ploop_balloon_clear_state(const char *device);
 PL_EXT int ploop_balloon_complete(const char *device);
-PL_EXT int ploop_balloon_check_and_repair(const char *device, char *mount_point, int repair);
-PL_EXT int ploop_discard(const char *device, const char *mount_point,
-		__u64 minblock, __u64 to_free);
+PL_EXT int ploop_balloon_check_and_repair(const char *device, const char *mount_point, int repair);
+PL_EXT int ploop_discard_get_stat_by_dev(const char *device, const char *mount_point,
+		struct ploop_discard_stat *pd_stat);
+PL_EXT int ploop_discard_by_dev(const char *device, const char *mount_point,
+		__u64 minlen_b, __u64 to_free, const int *stop);
 
 /* lock */
 int ploop_lock_di(struct ploop_disk_images_data *di);
@@ -296,21 +307,29 @@ int resize_gpt_partition(const char *devname);
 // misc
 void get_basedir(const char *fname, char *out, int len);
 __u32 crc32(const unsigned char *buf, unsigned long len);
-int ploop_is_on_nfs(const char *path);
 int store_statfs_info(const char *mnt, char *image);
+int drop_statfs_info(const char *image);
 int read_statfs_info(const char *image, struct ploop_info *info);
 int get_statfs_info(const char *mnt, struct ploop_info *info);
 int ploop_get_child_count_by_uuid(struct ploop_disk_images_data *di, const char *guid);
 int ploop_get_child_by_uuid(struct ploop_disk_images_data *di, const char *guid,  char **child_guid);
 int ploop_fname_cmp(const char *p1, const char *p2);
 int is_valid_guid(const char *guid);
-int read_line(const char *path, char *nbuf, int len);
+PL_EXT int read_line(const char *path, char *nbuf, int len);
+int read_line_quiet(const char *path, char *nbuf, int len);
 int is_valid_blocksize(__u32 blocksize);
+int run_prg(char *const argv[]);
+int guidcmp(const char *p1, const char *p2);
+int auto_mount_image(struct ploop_disk_images_data *di,
+		struct ploop_mount_param *param);
+void free_mount_param(struct ploop_mount_param *param);
+
 
 // merge
 PL_EXT int get_delta_info(const char *device, struct merge_info *info);
 PL_EXT int merge_image(const char *device, int start_level, int end_level, int raw, int merge_top,
 		char **images);
 int ploop_merge_snapshot_by_guid(struct ploop_disk_images_data *di, const char *guid, int merge_mode);
+int ploop_complete_running_operation(const char *device);
 
 #endif
