@@ -20,7 +20,6 @@
 #include <limits.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/param.h>
 #include <libxml/parser.h>
 #include <libxml/tree.h>
 #include <libxml/xmlwriter.h>
@@ -104,7 +103,7 @@ static int parse_ul(const char *str, __u64 *val)
 static int parse_xml(const char *basedir, xmlNode *root_node, struct ploop_disk_images_data *di)
 {
 	xmlNode *cur_node, *node;
-	char image[MAXPATHLEN];
+	char image[PATH_MAX];
 	const char *data = NULL;
 	const char *file = NULL;
 	const char *guid = NULL;
@@ -260,6 +259,28 @@ void get_basedir(const char *fname, char *out, int len)
 		*p = 0;
 }
 
+/* Convert to new format with constant TopGUID */
+static int convert_disk_descriptor(struct ploop_disk_images_data *di)
+{
+	if (di->top_guid != NULL && !guidcmp(di->top_guid, TOPDELTA_UUID))
+		return 0;
+
+	ploop_log(0, "DiskDescriptor.xml is in old format: converting");
+
+	if ((find_image_by_guid(di, TOPDELTA_UUID) != NULL) ||
+			(find_snapshot_by_guid(di, TOPDELTA_UUID) != -1)) {
+		ploop_err(0, "Can't convert: %s is in use",
+				TOPDELTA_UUID);
+		return -1;
+	}
+
+	ploop_log(0, "Changing %s to %s",
+			di->top_guid, TOPDELTA_UUID);
+	ploop_di_change_guid(di, di->top_guid, TOPDELTA_UUID);
+
+	return 0;
+}
+
 static int validate_disk_descriptor(struct ploop_disk_images_data *di)
 {
 	if (di->nimages == 0) {
@@ -275,19 +296,17 @@ static int validate_disk_descriptor(struct ploop_disk_images_data *di)
 	if (di->nimages != di->nsnapshots) {
 		int ret;
 
-		ret = ploop_add_snapshot_entry(di, BASE_UUID, NONE_UUID);
+		ret = ploop_add_snapshot_entry(di, TOPDELTA_UUID, NONE_UUID);
 		if (ret)
 			return ret;
 		if (di->top_guid == NULL)
-			di->top_guid = strdup(BASE_UUID);
+			di->top_guid = strdup(TOPDELTA_UUID);
 	}
-	if (di->top_guid == NULL && find_snapshot_by_guid(di, BASE_UUID) != -1) {
+	if (di->top_guid == NULL && find_snapshot_by_guid(di, TOPDELTA_UUID) != -1) {
 		/* Parallels VM compatibility.
 		 * The top delta is hardcoded {5fbaabe3-6958-40ff-92a7-860e329aab41}
 		 */
-		ploop_err(0, "No TopGuid found, using " BASE_UUID);
-		di->top_guid = strdup(BASE_UUID);
-		di->runtime->vm_compat = 1;
+		di->top_guid = strdup(TOPDELTA_UUID);
 	}
 	if (!is_valid_guid(di->top_guid)) {
 		ploop_err(0, "Validation of %s failed: invalid top delta %s",
@@ -302,11 +321,12 @@ static int validate_disk_descriptor(struct ploop_disk_images_data *di)
 	return 0;
 }
 
-int ploop_read_diskdescriptor(const char *fname, struct ploop_disk_images_data *di)
+static int read_diskdescriptor(const char *fname, struct ploop_disk_images_data *di)
 {
 	int ret;
-	char path[MAXPATHLEN];
-	char basedir[MAXPATHLEN];
+	char path[PATH_MAX];
+	char basedir[PATH_MAX];
+	struct stat st;
 	xmlDoc *doc = NULL;
 	xmlNode *root_element = NULL;
 
@@ -315,6 +335,12 @@ int ploop_read_diskdescriptor(const char *fname, struct ploop_disk_images_data *
 	if (di == NULL)
 		return -1;
 
+	/* workaround libxml2 SIGSEGV on empty document */
+	if (stat(fname, &st) || st.st_size == 0) {
+		ploop_err(0, "Can't parse %s", fname);
+		return -1;
+	}
+
 	if (realpath(fname, path) == NULL) {
 		ploop_err(errno, "Can't resolve %s", fname);
 		return -1;
@@ -335,11 +361,24 @@ int ploop_read_diskdescriptor(const char *fname, struct ploop_disk_images_data *
 		ret = validate_disk_descriptor(di);
 
 	xmlFreeDoc(doc);
-	xmlCleanupParser();
 
 	return ret;
 }
 
+int ploop_read_disk_descr(struct ploop_disk_images_data **di, const char *file)
+{
+	*di = alloc_diskdescriptor();
+	if (*di == NULL)
+		return SYSEXIT_NOMEM;
+
+	if (read_diskdescriptor(file, *di)) {
+		ploop_free_diskdescriptor(*di);
+		return SYSEXIT_DISKDESCR;
+	}
+
+	return 0;
+}
+
 static int normalize_image_name(const char *basedir, const char *image, char *out, int len)
 {
 	const char *p;
@@ -360,11 +399,14 @@ int ploop_store_diskdescriptor(const char *fname, struct ploop_disk_images_data
 	int i, rc = -1;
 	xmlTextWriterPtr writer = NULL;
 	xmlDocPtr doc = NULL;
-	char tmp[MAXPATHLEN];
-	char basedir[MAXPATHLEN] = "";
+	char tmp[PATH_MAX];
+	char basedir[PATH_MAX] = "";
 
 	ploop_log(0, "Storing %s", fname);
 
+	if (convert_disk_descriptor(di))
+		return -1;
+
 	get_basedir(fname, tmp, sizeof(tmp));
 
 	if (*tmp && realpath(tmp, basedir) == NULL) {
@@ -474,7 +516,7 @@ int ploop_store_diskdescriptor(const char *fname, struct ploop_disk_images_data
 		rc = xmlTextWriterWriteElement(writer, BAD_CAST "Preallocated",
 				BAD_CAST "1");
 		if (rc < 0) {
-			ploop_err(0, "Error at xmlTextWriter Blocksize");
+			ploop_err(0, "Error at xmlTextWriter Preallocated");
 			goto err;
 		}
 	}
@@ -537,6 +579,7 @@ int ploop_store_diskdescriptor(const char *fname, struct ploop_disk_images_data
 		ploop_err(0, "Error at xmlTextWriter Snapshots");
 		goto err;
 	}
+
 	if (di->top_guid != NULL) {
 		rc = xmlTextWriterWriteElement(writer, BAD_CAST "TopGUID",
 				BAD_CAST di->top_guid);
