@@ -7,7 +7,15 @@ while test -n "$1"; do
 		;;
 	   -i|--install)
 		build=yes
-		install=yes
+		install=i
+		;;
+	   -U|--update)
+		build=yes
+		install=U
+		;;
+	   -F|--freshen)
+		build=yes
+		install=F
 		;;
 	   -v|--verbose)
 		verbose=yes
@@ -59,7 +67,7 @@ read_spec
 SPEC_VR=$(echo $SPEC_VR | sed 's/-1$//')
 
 test "$build" = "yes" || exit 0
-make rpms
+make rpms || exit 1
 
-test "$install" = "yes" || exit 0
-sudo rpm -Uhv $(rpm --eval %{_rpmdir}/%{_arch})/ploop-*${GIT_VR}*.rpm
+test -z "$install" && exit 0
+sudo rpm -${install}hv $(rpm --eval %{_rpmdir}/%{_arch})/ploop-*${GIT_VR}*.rpm
